#ifndef MIE_ETAT_UTIL_H_
#define MIE_ETAT_UTIL_H_

#ifdef _WIN32
	#pragma warning(disable : 4799)
	#pragma warning(disable : 4127) // constatant condition
	#pragma warning(disable : 4996)
#endif

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <string>
#include "mie_common.h"
#include "etat_table.h"

namespace MIE_ETAT {

template<class T>
struct operatable {
	friend MIE_FORCE_INLINE T operator+(const T& a, const T& b) {
		T ret(a);
		ret += b;
		return ret;
	}
	friend MIE_FORCE_INLINE T operator-(const T& a, const T& b) {
		T ret(a);
		ret -= b;
		return ret;
	}
	friend MIE_FORCE_INLINE T operator*(const T& a, const T& b) {
		T ret(a);
		ret *= b;
		return ret;
	}
};

template<int n>
struct intpow3T { enum { val = intpow3T<n - 1>::val * 3 }; };

template<>
struct intpow3T<0> { enum { val = 1 }; };

static inline int intpow3(int x)
{
	int ret = 1;
	while (x > 0) {
		ret *= 3;
		x--;
	}
	return ret;
}
/*!
	xtbl([tblSize - 1] .. tbl[0])
	PȂoCi\bh
*/
template<class F>
static void power(F& x, const uint32_t *tbl, size_t tblSize)
{
	F t(x);
	x.set(1);
	for (size_t i = 0; i < tblSize; i++) {
		uint32_t n = tbl[i];
		for (int j = 0; j < 32; j++) {
			if (n & (1U << j)) {
				x *= t;
			}
			t *= t;
		}
	}
}

/**
	force to unroll loop
	if (step > 0) {
		for (int i = begin; i < end; i += step) p.calc<i>();
	} else {
		for (int i = begin; i > end; i += step) p.calc<i>();
	}
*/
template<class P, int begin, int end, int step = 1, int count = (step > 0 ? (end - begin + step - 1) / step : (begin - end - step - 1) / -step), int i = count>
struct Loop {
	MIE_FORCE_INLINE Loop(P& p)
	{
		p.template calc<begin + (count - i) * step>();
		Loop<P, begin, end, step, count, i - 1> run(p);
	}
};

template<class P, int begin, int end, int step, int count>
struct Loop<P, begin, end, step, count, 0> {
	Loop(P&) { }
};

/*
	clear array x[N]
*/
template<int N>
inline void rawClear(uint32_t (&x)[N], int n = N)
{
	::memset(x, 0, sizeof(*x) * n);
}
/*
	d[N] = s[N]
*/
template<int N>
inline void rawCopy(uint32_t (&d)[N], const uint32_t *s, int n = N)
{
	::memcpy(d, s, sizeof(*d) * n);
}

/*
	[dH:dL] += [sH:sL]
*/
template<typename T>
static MIE_FORCE_INLINE void rawAdd(T *dL, T *dH, T bL, T bH)
{
	T aL = *dL;
	T aH = *dH;
	T t = (aH | aL) & (bH | bL);
	*dL = t ^ (aL | bL);
	*dH = t ^ (aH | bH);
}

/*
	[dH:dL]  = [aH:aL] + [bH:bL]
*/
template<typename T>
static MIE_FORCE_INLINE void rawAdd(T *dL, T *dH, T aL, T aH, T bL, T bH)
{
	T t = (aH | aL) & (bH | bL);
	*dL = t ^ (aL | bL);
	*dH = t ^ (aH | bH);
}

/*
	[dH:dL] -= [sH:sL]
*/
template<typename T>
static MIE_FORCE_INLINE void rawSub(T *dL, T *dH, T bL, T bH)
{
	T aL = *dL;
	T aH = *dH;
	T t = (aH | aL) & (bH | bL);
	*dL = t ^ (aL | bH);
	*dH = t ^ (aH | bL);
}

template<typename T>
static MIE_FORCE_INLINE void rawSub(T *dL, T *dH, T aL, T aH, T bL, T bH)
{
	T t = (aH | aL) & (bH | bL);
	*dL = t ^ (aL | bH);
	*dH = t ^ (aH | bL);
}

static inline uint32_t rawGet(const uint32_t *L, const uint32_t *H, size_t pos)
{
	size_t q = pos / 32;
	size_t r = pos & 31;
	return ((L[q] >> r) & 1) | (((H[q] >> r) & 1) << 1);
}

static inline uint32_t GetMask(uint32_t len)
{
	assert(len < 32);
	return (1 << len) - 1;
}
/*
	return p[pos + len - 1, ..., pos + 0]
*/
static inline uint32_t rawGet(const uint32_t *p, uint32_t pos, uint32_t len = 32)
{
	assert(0 < len && len <= 32);
	size_t q = pos / 32;
	size_t r = pos & 31;
	uint32_t ret;
	if (r == 0) {
		ret = p[q];
	} else {
		ret = (p[q] >> r) + (p[q + 1] << (32 - r));
	}
	if (len == 32) {
		return ret;
	} else {
		return ret & GetMask(len);
	}
}

static inline void rawSet(uint32_t *L, uint32_t *H, size_t pos, uint32_t v)
{
	assert(v <= 2);
	size_t q = pos / 32;
	size_t r = pos & 31;
	L[q] &= ~(1 << r);
	H[q] &= ~(1 << r);
	L[q] |= (v & 1) << r;
	H[q] |= (v >> 1) << r;
}

/*
	non clear rawSet
*/
static inline void rawSetNC(uint32_t *L, uint32_t *H, size_t pos, uint32_t v)
{
	assert(v <= 2);
	size_t q = pos / 32;
	size_t r = pos & 31;
	L[q] |= (v & 1) << r;
	H[q] |= (v >> 1) << r;
}
/*
	[H:L] = -[H:L]
*/
template<typename T>
static inline void rawNeg(T *L, T *H, const T *sL, const T *sH)
{
	T t = *sH;
	*H = *sL;
	*L = t;
}

template<int N>
static std::string toString(const uint32_t (&x)[N])
{
	std::string ret;
	for (int i = 0; i < N; i++) {
		char buf[32];
		sprintf(buf, "%08x", x[N - 1 - i]);
		ret += buf;
		if (i < N - 1) ret += ':';
	}
	return ret;
}

template<int shift>
static MIE_FORCE_INLINE void expand32(uint32_t *out, uint32_t x)
{
	const uint32_t m10 = GetMask(10);
	const uint32_t m11 = GetMask(11);
	switch (shift) {
	case 0:
		out[0] = cubeTbl[x & m11];
		out[1] = cubeTbl[(x >> 11) & m11] << 1;
		out[2] = cubeTbl[x >> 22] << 2;
		break;
	case 1:
		out[0] = cubeTbl[x & m11] << 1;
		out[1] = cubeTbl[(x >> 11) & m10] << 2;
		out[2] = cubeTbl[x >> 21];
		break;
	case 2:
		out[0] = cubeTbl[x & m10] << 2;
		out[1] = cubeTbl[(x >> 10) & m11];
		out[2] = cubeTbl[x >> 21] << 1;
		break;
	default:
		MIE_NOT_HERE;
	}
}

template<int N>
static void put(const uint32_t (&x)[N], const char *msg = 0)
{
	if (msg) printf("%s ", msg);
	printf("%s\n", toString(x).c_str());
}

#define PUT(x) printf(#x "=%s\n", (x).toString().c_str())

} // end of namespace

#endif
