#define MIE_USE_ETAT_193
#define MIE_DONT_USE_MULTI_CORE // not multi thread
#define MIE_DONT_PRINT_SPACE_FOR_STR // not print space

#include "etat.h"
#include "etat_if.h"

using namespace MIE_ETAT;

#define NUM_OF_ARRAY(x) (sizeof(x) / sizeof(*x))

const size_t alignment = 16;
#ifdef _MSC_VER
	#include <malloc.h>
	static inline void *AlignedMalloc(size_t size)
	{
		return _aligned_malloc(size, alignment);
	}
	static inline void AlignedFree(void *ptr)
	{
		_aligned_free(ptr);
	}
#else
	#include <stdlib.h>
	static inline void *AlignedMalloc(size_t size)
	{
		void *p;
		int ret = posix_memalign(&p, alignment, size);
		return (ret == 0) ? p : 0;
	}
	static inline void AlignedFree(void *ptr)
	{
		free(ptr);
	}
#endif

const char *MIE_ErrToStr(MIE_Err err)
{
	static const char *tbl[16] = {
		"no err",
		"not support",
		"no memory",
		"no SSSE3",
		"small size",
	};
	if (err < 0 || err >= MIE_END_OF_ERR) {
		return "invalid err number";
	}
	return tbl[err];
}

#include "m193_if.h"

MIE_Err MIE_CreateSystem(MIE_System **psys, int m)
{
	if (!init()) return MIE_NO_SSSE3;

	switch (m) {
	case 193:
		return m193::createSystem(psys);
	default:
		return MIE_NOT_SUPPORT;
	}
}

void MIE_DestroySystem(MIE_System *sys)
{
	if (sys && sys->destroySystem) sys->destroySystem(sys);
}
