#ifndef MIE_ETAT_SRC_EXTENDED_F_H_
#define MIE_ETAT_SRC_EXTENDED_F_H_
/*
	extended class generated by F = F_3^m
	F3_3base : F^3
	F6base : F^6
	ECgeneral : E[F]
	CalcEtatPairing : pairing
*/
template<class F>
struct F3_3base : private operatable<F3_3base<F> > {
	enum {
		m = F::m,
		ma = F::ma
	};
	/*
		p^3 = p + b
		s^2 = -1
		a2p^2 + a1p + a0
	*/
	F a0, a1, a2;
	F3_3base()
	{
	}
	F3_3base(const F& _a0, const F& _a1, const F& _a2)
		: a0(_a0)
		, a1(_a1)
		, a2(_a2)
	{
	}
	explicit F3_3base(int x)
		: a0(x)
		, a1((uint32_t)0)
		, a2((uint32_t)0)
	{
	}
	void operator+=(const F3_3base& rhs)
	{
		a0 += rhs.a0;
		a1 += rhs.a1;
		a2 += rhs.a2;
	}
	void operator-=(const F3_3base& rhs)
	{
		a0 -= rhs.a0;
		a1 -= rhs.a1;
		a2 -= rhs.a2;
	}
	void operator*=(const F3_3base& rhs)
	{
		const F& b0 = rhs.a0;
		const F& b1 = rhs.a1;
		const F& b2 = rhs.a2;
		F t00, t11, t22, t01, t02, t12;
		F::mul(t00, a0, b0);
		F::mul(t11, a1, b1);
		F::mul(t22, a2, b2);

		F::add(t01, a0, a1);
		F::add(t02, b0, b1);
		t01 *= t02;
		F::add(t02, a0, a2);
		F::add(t12, b0, b2);
		t02 *= t12;
		F::add(t12, b1, b2);
		a2 += a1;
		t12 *= a2;

		t11 -= t00;
		F::add(a0, t11, t22);
		F::sub(a0, t12, a0);
		F::add(a1, t01, t12);
		F::add(a2, t02, t11);
		F::add(a1, a1, t11);
	}
	/*
		o1 = x1 * y
		o2 = x2 * y
	*/
	static inline void mulPara(F3_3base& o1, F3_3base& o2, const F3_3base& x1, const F3_3base& x2, const F3_3base& y)
	{
		const F& b0(y.a0);
		const F& b1(y.a1);
		const F& b2(y.a2);
		F c0, d0, c1, d1;
		F t;
		F t1_02, t2_02, t1_01, t2_01;
		F::add(t, b1, b2); // t = y1 + y2
		F::add(c0, x1.a1, x1.a2); // a1 + a2
		F::add(d0, x2.a1, x2.a2);
		F::addPara(t1_01, t1_02, x1.a1, x1.a2, x1.a0); // a1 + a0, a2 + a0
		F::addPara(t2_01, t2_02, x2.a1, x2.a2, x2.a0); // a1 + a0, a2 + a0
		F::mulPara(c0, d0, c0, d0, t); // (a1 + a2)(b1 + b2)

		F t0, t1;
		F::mulPara(t0, t1, x1.a1, x2.a1, b1); // a1b1
		F s0, s1;
		F::mulPara(s0, s1, x1.a0, x2.a0, b0); // a0b0
		t0 -= s0; // a1b1 - a0b0
		t1 -= s1;
		F::addsub(c1, c0, c0, t0);
		F::addsub(d1, d0, d0, t1);
		F::mulPara(o1.a0, o2.a0, x1.a2, x2.a2, b2); // a2b2
		F::sub(o1.a0, c0, o1.a0);
		F::sub(o2.a0, d0, o2.a0);

		F::add(t, b0, b1); // b0 + b1
		F::mulPara(s0, s1, t1_01, t2_01, t); // (a0 + a1)(b0 + b1)
		F::add(o1.a1, c1, s0);
		F::add(o2.a1, d1, s1);

		F::add(t, b0, b2);
		F::mulPara(o1.a2, o2.a2, t1_02, t2_02, t);
		o1.a2 += t0;
		o2.a2 += t1;
	}
	void square()
	{
		F t00, t11, t22, t01, t02, t12;
		F xTbl[XTBL_SIZE(g_pbit)];
		F::makePrevMulTbl<g_pbit>(xTbl, a0);
		F::mulLatter<g_pbit, g_w>(t00, xTbl, a0);
		F::mulLatter<g_pbit, g_w>(t01, xTbl, a1);
		F::mulLatter<g_pbit, g_w>(t02, xTbl, a2);

		F::makePrevMulTbl<g_pbit>(xTbl, a2);
		F::mulLatter<g_pbit, g_w>(t12, xTbl, a1);
		F::mulLatter<g_pbit, g_w>(t22, xTbl, a2);

		F::mul(t11, a1, a1);
		t12.neg();

		F::add(a0, t00, t12);
		F::sub(a1, t22, t01); a1 += t12;
		F::add(a2, t11, t22); a2 -= t02;
	}
	void cube()
	{
		a0.cube();
		a1.cube();
		a2.cube();
		a0 += a1;
		a0 += a2;
		a1 -= a2;
	}
	void inverse()
	{
		F t00, t11, t22, t01, t02, r;
		F::mul(t00, a0, a0);
		F::mul(t11, a1, a1);
		F::mul(t22, a2, a2);

		F::mulPara(t01, t02, a1, a2, a0);
		F t;
		F::add(t, a1, a2);
		t *= a1;
		a1 -= a0;
		a0 -= a2;
		a2 -= a1;
		a0 *= t00;
		a1 *= t11;
		a2 *= t22;
		F::add(r, a0, a1);
		r += a2;
		MIE_ETAT::inverse(r);
		t += t02;
		F::addPara(t00, t02, t00, t02, t22);
		t00 -= t;
		t22 -= t01;
		t11 -= t02;
	#if 1
		{ /* 8096 => 6800 */
			F xTbl[XTBL_SIZE(g_pbit)];
			F::template makePrevMulTbl<g_pbit>(xTbl, r);
			F::template mulLatter<g_pbit, g_w>(a0, xTbl, t00);
			F::template mulLatter<g_pbit, g_w>(a1, xTbl, t22);
			F::template mulLatter<g_pbit, g_w>(a2, xTbl, t11);
		}
	#else
		a0 = t00 * r;
		a1 = t22 * r;
		a2 = t11 * r;
	#endif
	}

	bool operator==(const F3_3base& rhs) const
	{
		return a0 == rhs.a0 && a1 == rhs.a1 && a2 == rhs.a2;
	}
	bool operator!=(const F3_3base& rhs) const { return !operator==(rhs); }
	bool isZero() const
	{
		return a0.isZero() && a1.isZero() && a2.isZero();
	}
	std::string toString() const
	{
		static const char *sep = "\n";
		return "(\n" + a0.toString() + sep + a1.toString() + sep + a2.toString() + ")";
	}
};

template<class F>
struct F6base : private operatable<F6base<F> > {
	enum {
		m = F::m,
		ma = F::ma
	};
	/*
		(a0, a1, a2, a3, a4, a5) x (1, p, p^2, s, sp, sp^2)
		x = a0 + a1p + a2p^2 + a3s + a4sp + a5sp^2
		  = (a0 + a1p + a2p^2) + (a3 + a4s + a5p^2)s
	*/
	F a0, a1, a2, a3, a4, a5, a6;
	F6base()
	{
	}
	F6base(const F& _a0, const F& _a1, const F& _a2, const F& _a3, const F& _a4, const F& _a5)
		: a0(_a0)
		, a1(_a1)
		, a2(_a2)
		, a3(_a3)
		, a4(_a4)
		, a5(_a5)
	{
	}
	void clear()
	{
		a0.clear();
		a1.clear();
		a2.clear();
		a3.clear();
		a4.clear();
		a5.clear();
	}
	void set(uint32_t x)
	{
		clear();
		a0.set(x);
	}
	void operator+=(const F6base& rhs)
	{
		a0 += rhs.a0; a1 += rhs.a1; a2 += rhs.a2; a3 += rhs.a3; a4 += rhs.a4; a5 += rhs.a5;
	}
	void operator-=(const F6base& rhs)
	{
		a0 -= rhs.a0; a1 -= rhs.a1; a2 -= rhs.a2; a3 -= rhs.a3; a4 -= rhs.a4; a5 -= rhs.a5;
	}
	/*
		multiply by (b0 = -u * u, b1 = -u, b2 = -1, b3, b4 = 0, b5 = 0)
	*/
	void mulBy(const F& u, const F& b3)
	{
		F b0;
		F::mul(b0, u, u);
		static const F _1(1);
		F r0, e0, e1, e2, e3, e4, e5, e6, e7, e8, e9, e10, e12, e13, e15, e18, e19, e22, e24;
		F::add(r0, a0, a2);
		F::addsub(e0, e12, r0, a1);
		F::addsub(r0, e9, _1, b0);
		r0.neg();
		F::addsub(e15, e3, r0, u);
		F::sub(r0, a0, a2);
		F::addsub(e18, e6, r0, a4);

		F::add(r0, a3, a5);
		F::addsub(e1, e13, r0, a4);
		F::sub(r0, a3, a5);
		F::addsub(e7, e19, r0, a1);
		F::addsub(e22, e10, b3, u);
		F::add(e24, a2, a5);
		F m0, m1, m2, m3, m4, m5, m6, m7, m8, m9, m10, m11;

		/*
			e0, e1, e2, e3, e5, e6, e7, e8, e9, e10, e11, e12, e13, e14, e15, e17, e18, e19, e20, e22, e23
			b3,
		*/
		F::mulPara(m2, m8, e1, e13, b3);
		F::mul(m0, e0, e3);
		F::mul(m5, e7, e10);

		F::mulPara(m3, m9, e6, e18, e9);
		F::mul(m11, e19, e22);
		F::mul(e2, e15, e12);

		F::add(m1, e0, e1);
		e3 += b3;
		m1 *= e3;
		m1.neg();
		F::add(m4, e9, e10);
		e7 += e6;
		m4 *= e7;
		e22 += e9;
		e19 += e18;
		F::mul(m10, e22, e19);

		/*
			29
			m0(3), m2(3), m3(2), m4(2), m5(3), m8(2), m9(2), m10(2), m11(2)
			a2(2)
			e2(2), e24(2)
		*/
		a2.neg();
		F::add(e0, m0, m4);
		e0 += a2;
		F::add(e1, m2, m10);
		e2 += a2;
		F::add(e4, e13, e12);
		e15 += b3;
		e4 *= e15;
		e4 -= e24;
		F::addsub(e7, e5, m2, m8);
		e7.neg();
		e7 += m5;
		e7 += m11;
		F::addsub(e8, e6, e2, m0);
		e8.neg();
		e8 += m3;
		e8 += m9;

		m4 += m10;
		m11 -= m3;
		m5 += e24; // e24
		m5 += m9; // m9
		/*
			20
			e0(2), e1(2), e4(2), e5(2), e6(2), e7(2), e8(2)
			m1(3), m4(1), m5(1), m11(1)
		*/
		F::addsub(a3, a0, e1, e0); // e0, e1
		a0 += m11; // m11
		a3 += m5; // m5
		F::addsub(a1, a4, e5, e6); // e5, e6
		F::addsub(a5, a2, e7, e8); // e7, e8
		F::addPara(a4, a5, a4, a5, e4);
		F::addPara(a3, a4, a3, a4, m1);
		m4 += m1; // m1
		a5 -= m4; // m4
	}
	/*
		15M + 67A
	*/
	void operator*=(const F6base& rhs)
	{
		const F& b0 = rhs.a0;
		const F& b1 = rhs.a1;
		const F& b2 = rhs.a2;
		const F& b3 = rhs.a3;
		const F& b4 = rhs.a4;
		const F& b5 = rhs.a5;
		F r0, e0, e1, e2, e3, e4, e5, e6, e7, e8, e9, e10, e11, e12, e13, e14, e15, e16, e17, e18, e19, e20, e21, e22, e23, e24, e25;
#if 1
		F::add(r0, a0, a2);
		F::add(e0, r0, a1);
		F::sub(e12, r0, a1);
		F::add(r0, b0, b2);
		F::add(e3, r0, b1);
		F::sub(e15, r0, b1);
		F::sub(r0, a0, a2);
		F::add(e18, r0, a4);
		F::sub(e6, r0, a4);
		F::sub(r0, b0, b2);
		F::add(e21, r0, b4);
		F::sub(e9, r0, b4);
		F::add(r0, a3, a5);
		F::add(e1, r0, a4);
		F::sub(e13, r0, a4);
		F::add(r0, b3, b5);
		F::add(e4, r0, b4);
		F::sub(e16, r0, b4);
		F::sub(r0, a3, a5);
		F::add(e7, r0, a1);
		F::sub(e19, r0, a1);
		F::sub(r0, b3, b5);
		F::add(e10, r0, b1);
		F::sub(e22, r0, b1);
#else
		F::add(r0, a0, a2);
		F::addsub(e0, e12, r0, a1);
		F::add(r0, b0, b2);
		F::addsub(e3, e15, r0, b1);
		F::sub(r0, a0, a2);
		F::addsub(e18, e6, r0, a4);
		F::sub(r0, b0, b2);
		F::addsub(e21, e9, r0, b4);
		F::add(r0, a3, a5);
		F::addsub(e1, e13, r0, a4);
		F::add(r0, b3, b5);
		F::addsub(e4, e16, r0, b4);
		F::sub(r0, a3, a5);
		F::addsub(e7, e19, r0, a1);
		F::sub(r0, b3, b5);
		F::addsub(e10, e22, r0, b1);
#endif
		F::add(e2, e0, e1); F::add(e5, e3, e4); F::add(e8, e6, e7);
		F::add(e11, e9, e10);
		F::add(e14, e12, e13); F::add(e17, e15, e16); F::add(e20, e18, e19); F::add(e23, e21, e22);
		F::add(e24, a2, a5); F::add(e25, b2, b5);
		F m0, m1, m2, m3, m4, m5, m6, m7, m8, m9, m10, m11, m12, m13, m14;
		F::mul(m0, e0, e3); F::mul(m1, e2, e5); F::mul(m2, e1, e4);
		F::mul(m3, e6, e9); F::mul(m4, e8, e11); F::mul(m5, e7, e10);
		F::mul(m6, e12, e15); F::mul(m7, e14, e17); F::mul(m8, e13, e16);
		F::mul(m9, e18, e21); F::mul(m10, e20, e23); F::mul(m11, e19, e22);
		F::mul(m12, a2, b2); F::mul(m13, e24, e25); F::mul(m14, a5, b5);
		F::add(e0, m0, m4); e0 += m12; F::add(e1, m2, m10); e1 += m14;
		F::add(e2, m6, m12); F::add(e3, m8, m14); e3.neg(); F::add(e4, m7, m13);
		F::addsub(e5, e7, e3, m2);
		e7 += m5;
		e7 += m11;
		F::addsub(e8, e6, e2, m0);
		F::addsub(a3, a0, e1, e0);
		a0 += m11;
		a0 -= m3;
		m3 += m9;
		e8 -= m3;
		m13 += m1;
		m5 += m9;
		a3 += m5;
		a3 -= m13;
		F::addsub(a1, a4, e5, e6);
		F::addsub(a2, a5, e7, e8);
		F::addsub(e7, e8, e4, m1);
		a4 += e8;
		a5 += e7;
		m4 += m10;
		a5 -= m4;
	}
	void cube()
	{
		F::cube(a0);
		F::cube(a1);
		F::cube(a2);
		F::cube(a3);
		F::cube(a4);
		F::cube(a5);
		a1 += a2;
		a0 += a1;
		a1 += a2;
		a3 += a4; a3 += a5; a3.neg();
		F::sub(a4, a5, a4);
		a5.neg();
	}
	void cube(uint32_t n)
	{
		while (n > 0) {
			cube();
			n--;
		}
	}
	void cubeRoot()
	{
		F::cubeRoot(a0);
		F::cubeRoot(a1);
		F::cubeRoot(a2);
		F::cubeRoot(a3);
		F::cubeRoot(a4);
		F::cubeRoot(a5);
		a0 -= a1;
		F::addPara(a0, a1, a0, a1, a2);
		a3 += a5;
		a3 -= a4;
		a3.neg();
		a4 += a5; a4.neg();
		a5.neg();
	}
	/* calc x^(3^m + 1) */
	void lambda()
	{
		static const F one(1);
		F v0, v1, v2, v3, v4, v5, v6, v7, v8;

		F::mulPara(v0, v2, a0, a1, a2);
		F::mulPara(v1, v3, a3, a4, a5);
		F::mul(v5, a3, a1);
		F::mul(v6, a0, a4);
		a0 += a3;
		a2 -= a5;
		a1 += a4;

		F::mulPara(v4, v8, a0, a1, a2);
		F::mul(v7, a0, a1);
		F::addsub(v2, v3, v3, v2);
		F::addsub(v5, v6, v5, v6);
		F::addsub(v0, v1, v1, v0);
		v7 -= v5;
		F::sub(a0, v0, v2);
		a0 += one;
		if ((m % 12) == 1) {
			F::sub(a1, v7, v2);
		} else {
			F::neg(a1, v7);
		}
		F::add(a5, v3, v8);
		F::add(a2, v2, v7);
		v1 += v4;
		F::add(a3, v6, v1);
		F::sub(a4, a5, v1);
	}
	bool operator==(const F6base& rhs) const
	{
		return a0 == rhs.a0 && a1 == rhs.a1 && a2 == rhs.a2 && a3 == rhs.a3 && a4 == rhs.a4 && a5 == rhs.a5;
	}
	bool operator!=(const F6base& rhs) const { return !(*this == rhs); }
	bool isZero() const
	{
		return a0.isZero() && a1.isZero() && a2.isZero() && a3.isZero() && a4.isZero() && a5.isZero();
	}
	std::string toString() const
	{
		static const char *sep = "\n";
		return "(\n" + a0.toString() + sep + a1.toString() + sep + a2.toString() + sep + a3.toString() + sep + a4.toString() + sep + a5.toString() + ")";
	}
	static inline void mulTri(F3_3base<F>& ff, F3_3base<F>& f0, F3_3base<F>& f1, const F6base& in)
	{
		const F& a0(in.a0);
		const F& a1(in.a1);
		const F& a2(in.a2);
		const F& a3(in.a3);
		const F& a4(in.a4);
		const F& a5(in.a5);
		F a00, a01, a02, a03, a04, a05;
		F a11, a12, a13, a14, a15;
		F a22, a23, a24, a25;
		F a33, a34, a35;
		F a44, a45;
		F a55;
		F xTbl[XTBL_SIZE(g_pbit)];
		F::template makePrevMulTbl<g_pbit>(xTbl, a0);
		F::template mulLatter<g_pbit, g_w>(a00, xTbl, a0);
		F::template mulLatter<g_pbit, g_w>(a01, xTbl, a1);
		F::template mulLatter<g_pbit, g_w>(a02, xTbl, a2);
		F::template mulLatter<g_pbit, g_w>(a03, xTbl, a3);
		F::template mulLatter<g_pbit, g_w>(a04, xTbl, a4);
		F::template mulLatter<g_pbit, g_w>(a05, xTbl, a5);
		F::template makePrevMulTbl<g_pbit>(xTbl, a1);
		F::template mulLatter<g_pbit, g_w>(a11, xTbl, a1);
		F::template mulLatter<g_pbit, g_w>(a12, xTbl, a2);
		F::template mulLatter<g_pbit, g_w>(a13, xTbl, a3);
		F::template mulLatter<g_pbit, g_w>(a14, xTbl, a4);
		F::template mulLatter<g_pbit, g_w>(a15, xTbl, a5);
		F::template makePrevMulTbl<g_pbit>(xTbl, a2);
		F::template mulLatter<g_pbit, g_w>(a22, xTbl, a2);
		F::template mulLatter<g_pbit, g_w>(a23, xTbl, a3);
		F::template mulLatter<g_pbit, g_w>(a24, xTbl, a4);
		F::template mulLatter<g_pbit, g_w>(a25, xTbl, a5);
		F::template makePrevMulTbl<g_pbit>(xTbl, a3);
		F::template mulLatter<g_pbit, g_w>(a33, xTbl, a3);
		F::template mulLatter<g_pbit, g_w>(a34, xTbl, a4);
		F::template mulLatter<g_pbit, g_w>(a35, xTbl, a5);
		F::template makePrevMulTbl<g_pbit>(xTbl, a4);
		F::template mulLatter<g_pbit, g_w>(a44, xTbl, a4);
		F::template mulLatter<g_pbit, g_w>(a45, xTbl, a5);
		F::mul(a55, a5, a5);
		a12.neg();
		F::add(f0.a0, a00, a12);
		F::sub(f0.a1, a22, a01); f0.a1 += a12;
		F::add(f0.a2, a11, a22); f0.a2 -= a02;
		a45.neg();
		F::add(f1.a0, a33, a45);
		F::sub(f1.a1, a55, a34); f1.a1 += a45;
		F::add(f1.a2, a44, a55); f1.a2 -= a35;

		F::add(ff.a0, a15, a24);
		F::add(ff.a1, a04, a13);
		F::addPara(ff.a0, ff.a1, a03, ff.a1, ff.a0);
		F::add(ff.a2, a05, a14); ff.a2 += a23;
		F::addPara(ff.a1, ff.a2, ff.a1, ff.a2, a25);
	}
	static inline void power3_3m_1(F6base& out, const F6base& in)
	{
		F3_3base<F> f0(in.a0, in.a1, in.a2), f1(in.a3, in.a4, in.a5), ff, r;
#if 1
		mulTri(ff, f0, f1, in);
#else
		ff = f0 * f1;
		f0.square();
		f1.square();
#endif
		r = f0;
		r += f1;
		r.inverse();
		f0 -= f1;
		// out = [f0 * r, ff * r]
		F3_3base<F>::mulPara(((F3_3base<F>*)&out)[0], ((F3_3base<F>*)&out)[1], f0, ff, r);
	}
	void finalExponentiation()
	{
		/*
			(3^(3m) - 1)(3^m + 1)(3^m - 3^((m+1)/2) + 1)
			=(3^(3m) - 1)((3^((m-1)/2) - 1) 3^((m+1)/2) + 1)(3^m + 1)
			= a * (b * 3^c + 1)(3^m + 1)
			a = 3^(3m) - 1 = 0x25609213623D7D6219085CF49D306450BF6519835586C19D3A4F3A2C5F35B44A300C8A76E11708B5495B9C3EE756BBF19E3FD15CE625D3C0539A;
			b = 3^((m-1)/2) - 1 = 0x10E425C56DAFFABC35C0;
			c = (m+1)/2
		*/
		F6base k, g;
		power3_3m_1(k, *this); /* k  k^(3^(3m) - 1) */
		k.lambda(); /* k  k^(3^m + 1) */
		g = k;
		k.lambda(); /* k  k^(3^m + 1) */
		g.cube((m + 1) / 2);
//		g.cube_mP1D2(); // g.cube((m + 1) / 2);
		g.a3.neg();
		g.a4.neg();
		g.a5.neg();
		*this = k * g;
	}
};

/*
	y^2 = x^3 - x + b
*/
template<class F>
struct ECgeneral : private operatable<ECgeneral<F> > {
	const F& one() const { static const F _1(b); return _1; }
	enum {
		b = 1
	};
	F x, y;
	/*
		set (x, y) for given _x
		@param _x [in] candidate of x
		@note x = x or -x
		@note y.get(0) is always 0 or 1
	*/
	bool set(const F& _x, bool tryAgain = false)
	{
		F t(_x);
		F::cube(t);
		t -= _x;
		t += one();
		y = t;
		if (F::squareRoot(y)) {
			x = _x;
			if (y.get(0) == 2) {
				y.neg();
			}
			return true;
		} else if (tryAgain) {
			t = _x;
			t.neg();
			return set(t, false);
		} else {
			return false;
		}
	}
	ECgeneral& operator+=(const ECgeneral& rhs)
	{
		if (x != rhs.x) {
			F s, t;
			F r = rhs.x - x;
			inverse(r);
			s = (rhs.y - y) * r;
			t = (y * rhs.x - rhs.y * x) * r;
			x.neg(); x -= rhs.x; x += s * s;
			y = s * x; y += t; y.neg();
			return *this;
		} else {
			/*
				2P = (x + 1 / y^2, -y - 1 / y^3)
			*/
			F r = y;
			inverse(r);
			x += r * r;
			y.neg();
			F::cube(r);
			y -= r;
			return *this;
		}
	}
	bool isValid() const
	{
		F t(x);
		F::cube(t);
		t += one(); t -= x; t -= y * y;
		return t.isZero();
	}
	void cube()
	{
		/* 3P = (x^9 - b, -y^9) */
		F::cube(x);
		F::cube(x);
		x -= one();
		F::cube(y);
		F::cube(y);
		y.neg();
	}
	void neg()
	{
		y.neg();
	}
	bool operator==(const ECgeneral& rhs) const
	{
		return x == rhs.x &&  y == rhs.y;
	}
	std::string toString() const
	{
		static const char *sep = "\n";
		return "(\n" + x.toString() + sep + y.toString() + ")";
	}
};

namespace local {

template<class F>
static inline void cubeRootSeq(F *tbl)
{
	/* x^(3^m) = x, then x^(1/3^i) = x^(3^(m - i)) */
	F t = tbl[0];
#ifdef MIE_NEW_VERSION
	const int n = F::lpN + 1;
#else
	const int n = F::lpN;
#endif
	F::cubeN(t, n);
	tbl[n - 1] = t;
	for (size_t i = n - 2; i >= 1; i--) {
		F::cube(t);
		tbl[i] = t;
	}
}

template<class F>
struct MakeCubeSeq {
	F *p_, *q_;
	MakeCubeSeq(F *p, F *q)
		: p_(p), q_(q)
	{
	}
	static void calc(MakeCubeSeq<F> *self)
	{
		cubeRootSeq(self->p_);
		F::cubeSeq(self->q_);
	}
};

template<class F>
struct MulHalf {
	F6base<F> *f_;
	const F *xp_, *xq_, *yp_, *yq_;
	size_t pos_;
	MulHalf(F6base<F> *f, const F *xp, const F *xq, const F *yp, const F *yq, size_t pos)
		: f_(f), xp_(xp), xq_(xq), yp_(yp), yq_(yq), pos_(pos)
	{
	}
	static void calc(MulHalf<F> *self)
	{
		static const F one(1);
		size_t pos = self->pos_;
		const F *xp(self->xp_ + pos), *xq(self->xq_ + pos), *yp(self->yp_ + pos), *yq(self->yq_ + pos);
		F6base<F> *f(self->f_);
#if 1
		F xTbl[XTBL_SIZE(g_pbit)];
		for (size_t i = 0; i < F::lpN / 4; i++) {
			F6base<F> h;
			/* 5.2Kclk(mulPara x 5 + mul x 5) */
			F u, up, v, vp;
			F::add(u, xp[i * 2], xq[i * 2]);
			F::mul(v, yp[i * 2], yq[i * 2]);
			F::add(up, xp[i * 2 + 1], xq[i * 2 + 1]);
			F::mul(vp, yp[i * 2 + 1], yq[i * 2 + 1]);
			F::addPara(u, up, u, up, one);
			F uup, vvp;
			F::mul(uup, u, up);
			F::mul(vvp, v, vp);
			F::add(h.a4, uup, vvp);
			F::add(h.a5, v, vp);
			v += u;
			vp += up;
			v *= vp;
			h.a4 -= v;
			u += up;
#if 1
			F::template makePrevMulTbl<g_pbit>(xTbl, u);
			F::template mulLatter<g_pbit, g_w>(h.a3, xTbl, h.a4);
			F::template mulLatter<g_pbit, g_w>(h.a1, xTbl, uup); h.a1 += u;
			F::template mulLatter<g_pbit, g_w>(h.a2, xTbl, u); h.a2 -= uup;
			F::addPara(h.a1, h.a2, h.a1, h.a2, one);
			vvp -= u;
			F::template makePrevMulTbl<g_pbit>(xTbl, uup);
			F::template mulLatter<g_pbit, g_w>(u, xTbl, h.a5); h.a3 += u;
			F::template mulLatter<g_pbit, g_w>(h.a0, xTbl, uup); h.a0 -= vvp;
#else
			h.a3 = h.a4 * u + h.a5 * uup;
			h.a0 = u - vvp + uup * uup;
			h.a1 = uup * u + (u + one);
			h.a2 = u * u - uup + one;
#endif
			h.a5.neg();
			/* 8.3Kclk(mul x 15)(is it fast if using FFT?)  */
			*f *= h;
		}
#else
		for (size_t i = 0; i < F::lpN / 2; i++) {
			F u;
			F::add(u, xp[i], xq[i]);
			u += one;
			/* 7.5Kclk(mulPara x 2 + mul x 13) */
			f->mulBy(u, yp[i] * yq[i]);
		}
#endif
	}
};

/*
	6M + 21A
	p^3 = p + 1, s^2 = -1
	out = (a0 + b0s + c0p + p^2)(a1 + b1s + c1p + p^2)
*/
template<class F>
static void multi6M(F6base<F>& out, const F& a0, const F& b0, const F& c0, const F& a1, const F& b1, const F& c1)
{
	static const F one(1);
	F aa, bb, cc, ab, bc, ac, sumc;
	F::mul(aa, a0, a1); // aa = a0 a1
	F::mul(bb, b0, b1); // bb = b0 b1
	F::mul(cc, c0, c1); // cc = c0 c1
	F::add(out.a5, b0, b1); // b0 + b1
	F::add(sumc, c0, c1);

	F::add(out.a0, aa, sumc);
	out.a0 -= bb; // (a0 a1 - b0 b1 + c0 + c1)

	/*
		(a0 + c0)(a1 + c1) - a0 a1 - c0 c1 + (c0 + c1) + 1
		= (a0 c1 + a1 c0) + (c0 + c1) + 1
	*/
	F::mul(out.a1, a0 + c0, a1 + c1);
	out.a1 -= aa;
	out.a1 -= cc;
	out.a1 += sumc;
	out.a1 += one;
	/*
		(a0 + a1) + c0 c1 + 1
	*/
	F::add(out.a2, a0 + a1, cc);
	out.a2 += one;
	/*
		(a0 + b0)(a1 + b1) - a0 a1 - b0 b1
		= a0 b1 + a1 b0
	*/
	F::mul(out.a3, a0 + b0, a1 + b1);
	out.a3 -= aa;
	out.a3 -= bb;
	/*
		(b0 + c0)(b1 + c1) - b0 b1 - c0 c1
		= b0 c1 + b1 c0
	*/
	F::mul(out.a4, b0 + c0, b1 + c1);
	out.a4 -= bb;
	out.a4 -= cc;
}

} // end of namespace

template<class F>
static F6base<F> CalcEtatPairing(const ECgeneral<F>& P, const ECgeneral<F>& Q)
{
	static const F one(1);
	const size_t n = F::lpN;
#ifdef MIE_NEW_VERSION
	F xp[n + 1], xq[n], yp[n + 1], yq[n];
	xp[0] = P.x;
	xq[0] = Q.x;
	yp[0] = P.y;
	yq[0] = Q.y;
	if (ECgeneral<F>::b == 1) {
		yp[0].neg();
	}
#ifdef MIE_USE_MULTI_CORE
	{
		local::MakeCubeSeq<F> p1(xp, xq);
		local::MakeCubeSeq<F> p2(yp, yq);
		s_parallelizer.calc((void*)&p1, (void*)&p2, (void(*)(void*))&local::MakeCubeSeq<F>::calc);
	}
#else
	local::cubeRootSeq(xp); // 12.4Kclk -> 8.5Kclk
	local::cubeRootSeq(yp);
	F::cubeSeq(xq); // 4.5Kclk -> 4.1Kclk
	F::cubeSeq(yq);
#endif
	F6base<F> R, S1, S2, g, h;

	F t, u;
	t = xp[1] + xq[0];

	R.clear();
	R.a0 = yp[1] * t; R.a0.neg();
	R.a1 = yp[1];
	R.a3 = yq[0];
	g.clear();
	g.a0 = t * t; g.a0.neg();
	g.a1 = t; g.a1.neg();
	g.a2 = one; g.a2.neg();
	g.a3 = yp[1] * yq[0];
	R *= g;

	for (int i = 0; i < (F::m - 1) / 4; i++) {
		F t0, t1, u0, u1;
		t0 = xp[i * 2 + 2] + xq[i * 2 + 1];
		u0 = yp[i * 2 + 2] * yq[i * 2 + 1];
		t1 = xp[i * 2 + 3] + xq[i * 2 + 2];
		u1 = yp[i * 2 + 3] * yq[i * 2 + 2];
#if 1
		u0.neg();
		u1.neg();
		local::multi6M(S1, t0 * t0, u0, t0, t1 * t1, u1, t1); // S1 *= S1;
#else
		S1.clear();
		S1.a0 = t0 * t0; S1.a0.neg();
		S1.a1 = t0; S1.a1.neg();
		S1.a2 = one; S1.a2.neg();
		S1.a3 = u0;
		S2.clear();
		S2.a0 = t1 * t1; S2.a0.neg();
		S2.a1 = t1; S2.a1.neg();
		S2.a2 = one; S2.a2.neg();
		S2.a3 = u1;
		S1 *= S2;
#endif
		R *= S1;
	}
	/* 75Kclk */
	R.finalExponentiation();
	return R;

#else ///////////////////////////////////////////////

#if defined(MIE_USE_MULTI_CORE) && !defined(MIE_USE_MULTI_CORE_WITH_BUSY_LOOP)
	s_parallelizer.wakeup();
#endif
	F xp[n], xq[n], yp[n], yq[n];
	xp[0] = P.x;
	xq[0] = Q.x;
	yp[0] = P.y;
	yq[0] = Q.y;
	if (ECgeneral<F>::b == 1) {
		yp[0].neg();
	}
	F6base<F> f;
	F::add(f.a0, xp[0], xq[0]);
	f.a0 += one;
	f.a0 *= yp[0];
	f.a0.neg();
	f.a1 = yp[0];
	f.a2.clear();
	f.a3 = yq[0];
	f.a4.clear();
	f.a5.clear(); /* f = -yp(xp + xq + 1) + ypp + yqs */
	/* 24.5K(single) -> 13.0K(multi) */
#ifdef MIE_USE_MULTI_CORE
	{
		local::MakeCubeSeq<F> p1(xp, xq);
		local::MakeCubeSeq<F> p2(yp, yq);
		s_parallelizer.calc((void*)&p1, (void*)&p2, (void(*)(void*))&local::MakeCubeSeq<F>::calc);
	}
#else
	local::cubeRootSeq(xp); // 12.4Kclk -> 8.5Kclk
	local::cubeRootSeq(yp);
	F::cubeSeq(xq); // 4.5Kclk -> 4.1Kclk
	F::cubeSeq(yq);
#endif
	F6base<F> h;
	h.a2 = one; h.a2.neg();
	F::add(h.a1, xp[n / 2], xq[n / 2]);
	h.a1 += one;
	F::mul(h.a0, h.a1, h.a1);
	h.a0.neg();
	h.a1.neg();
	F::mul(h.a3, yp[n / 2], yq[n / 2]);
	h.a4.clear();
	h.a5.clear();
	/* 310K(single)  156K(multi) */
	{
		local::MulHalf<F> p1(&f, xp, xq, yp, yq, 0);
		local::MulHalf<F> p2(&h, xp, xq, yp, yq, n / 2 + 1);
#ifdef MIE_USE_MULTI_CORE
		s_parallelizer.calc((void*)&p1, (void*)&p2, (void(*)(void*))&local::MulHalf<F>::calc);
#else
		local::MulHalf<F>::calc(&p1);
		local::MulHalf<F>::calc(&p2);
#endif
	}
#if defined(MIE_USE_MULTI_CORE) && !defined(MIE_USE_MULTI_CORE_WITH_BUSY_LOOP)
	s_parallelizer.suspend();
#endif
	f *= h; // 8K
	/* 75Kclk */
	f.finalExponentiation();
	return f;
#endif
}

#endif
