#ifndef MIE_MIE_SRC_F3BASE193_H_
#define MIE_MIE_SRC_F3BASE193_H_

typedef MIE_ETAT::F3base<193, 12> F3_193;

template<>
MIE_FORCE_INLINE F3_193::F3base(const F3base& rhs)
{
	V256(rhs.L_).store(L_);
	V256(rhs.H_).store(H_);
}

template<>
MIE_FORCE_INLINE void F3_193::operator=(const F3base& rhs)
{
	V256(rhs.L_).store(L_);
	V256(rhs.H_).store(H_);
}

template<>
MIE_FORCE_INLINE void F3_193::clear()
{
	V128 x = Zero();
	x.store(L_); x.store(L_ + 4);
	x.store(H_); x.store(H_ + 4);
}

template<>
MIE_FORCE_INLINE void F3_193::swap(F3base& x)
{
	swap256(L_, x.L_);
	swap256(H_, x.H_);
}

template<>
MIE_FORCE_INLINE void F3_193::set(const uint32_t *L, const uint32_t *H)
{
	static const MIE_ALIGN(16) uint32_t mask[4] = { 0xffffffff, 0xffffffff, 1, 0 };
	V256 t(L);
	t.x1_ &= V128(mask);
	t.store(L_);
	t.load(H);
	t.x1_ &= V128(mask);
	t.store(H_);
}

template<>
MIE_FORCE_INLINE void F3_193::neg(F3base& out, const F3base& a)
{
	V256 t(a.L_);
	V256(a.H_).store(out.L_);
	t.store(out.H_);
}

template<>
MIE_FORCE_INLINE int F3_193::degree(const F3base& x)
{
	if (x.L_[N - 1] | x.H_[N - 1]) return m - 1;
	for (int i = N - 2; i >= 0; i--) {
		int v = x.L_[i] | x.H_[i];
		int d;
		if (mie_bsr(&d, v)) {
			return d + i * 32;
		}
	}
	return 0;
}

template<>
MIE_FORCE_INLINE void F3_193::add(F3base& out, const F3base& a, const F3base& b)
{
	V256 aL(a.L_), aH(a.H_), bL(b.L_), bH(b.H_);
#ifdef MIE_USE_6OP_ADD
	bL ^= aL;
	bH ^= aH;
	aL ^= bH;
	aH ^= bL;
	andn(bH, aH).store(out.L_);
	andn(bL, aL).store(out.H_);
#else
	V256 t((aL | aH) & (bL | bH));
	(t ^ (aL | bL)).store(out.L_);
	(t ^ (aH | bH)).store(out.H_);
#endif
}

template<>
MIE_FORCE_INLINE void F3_193::sub(F3base& out, const F3base& a, const F3base& b)
{
	V256 aL(a.L_), aH(a.H_), bL(b.H_), bH(b.L_);
#ifdef MIE_USE_6OP_ADD
	bL ^= aL;
	bH ^= aH;
	aL ^= bH;
	aH ^= bL;
	andn(bH, aH).store(out.L_);
	andn(bL, aL).store(out.H_);
#else
	V256 t((aL | aH) & (bL | bH));
	(t ^ (aL | bL)).store(out.L_);
	(t ^ (aH | bH)).store(out.H_);
#endif
}

template<>
void F3_193::addsub(F3base& addOut, F3base& subOut, const F3base& a, const F3base& b)
{
	V256 bL(b.L_), bH(b.H_), tb(bH | bL), aL(a.L_), aH(a.H_), t((aL | aH) & tb);
	((aL | bL) ^ t).store(addOut.L_);
	((aL | bH) ^ t).store(subOut.L_);
	((aH | bH) ^ t).store(addOut.H_);
	((aH | bL) ^ t).store(subOut.H_);
}

template<>
void F3_193::addPara(F3base& out1, F3base& out2, const F3base& a1, const F3base& a2, const F3base& b)
{
	V256 bL(b.L_), bH(b.H_), tb(bH | bL), aL(a1.L_), aH(a1.H_), t((aL | aH) & tb);
	((aL | bL) ^ t).store(out1.L_);
	((aH | bH) ^ t).store(out1.H_);
	aL.load(a2.L_);
	aH.load(a2.H_);
	t = (aL | aH) & tb;
	((aL | bL) ^ t).store(out2.L_);
	((aH | bH) ^ t).store(out2.H_);
}

template<>
MIE_FORCE_INLINE bool F3_193::isZero() const
{
	return (V128(L_) | V128(H_) | V128(L_ + 4) | V128(H_ + 4)).cmpeq(Zero()).pmovmskb() == 0xffff;
}

template<>
MIE_FORCE_INLINE bool F3_193::operator==(const F3base& rhs) const
{
	V256 L(L_), H(H_);
	L.cmpeq(V256(rhs.L_));
	H.cmpeq(V256(rhs.H_));
	L &= H;
	return (L.x0_ & L.x1_).pmovmskb() == 0xffff;
}

/*
	[x2:x1:x0] << bit
*/
template<int bit>
MIE_FORCE_INLINE void shlBitNotRet(V128& x0, V128& x1, V128& x2)
{
	assert(2 <= bit && bit < 8);
	V128 t(x2);
	x2 = psrlq<8 - bit>(palignr<15>(x2, x1)) | psllq<bit>(x2);
	x1 = psrlq<8 - bit>(palignr<15>(x1, x0)) | psllq<bit>(x1);
	x0 = psllq<bit>(x0) | psrlq<8 - bit>(pslldq<1>(x0));
}
/*
	[x2:x1:x0] << bit
	return x2 >> (128 - bit)
*/
template<int bit>
MIE_FORCE_INLINE uint32_t shlBit(V128& x0, V128& x1, V128& x2)
{
	uint32_t ret = psrldq<12>(x2).get32bit();
	shlBitNotRet<bit>(x0, x1, x2);
	return ret >> (32 - bit);
}

/*
	rotlByte<8>(x0, x1, x2)
	shlBitNotRet(x0, x1, x2)
*/
template<int bit>
MIE_FORCE_INLINE void rotAndShlBitNotRet(V128& x0, V128& x1, V128& x2)
{
	V128 t(x2);
	x2 = psllq<bit>(palignr<8>(x2, x1)) | psrlq<64 - bit>(x1);
	x1 = psllq<bit>(palignr<8>(x1, x0)) | psrlq<64 - bit>(x0);
	x0 = psllq<bit>(palignr<8>(x0, t)) | psrlq<64 - bit>(punpckhqdq(Zero(), t));
}

template<int bit>
MIE_FORCE_INLINE uint32_t rotAndShlBit(V128& x0, V128& x1, V128& x2)
{
	uint32_t ret = psrlq<64 - bit>(x2).get32bit();
	rotAndShlBitNotRet<bit>(x0, x1, x2);
	return ret;
}

#define ADD_VEC(aL, aH, bL, bH) bL ^= aL; bH ^= aH; aL ^= bH; aH ^= bL; bH = andn(bH, aH); aH = andn(bL, aL); aL = bH

template<>
void MIE_FORCE_INLINE F3baseDbl<F3_193>::reduce(F3_193& ret)
{
	static const MIE_ALIGN(16) uint32_t mask[4] = { 0xffffffff, 0xffffffff, 1, 0 };
	V256 yL, yH;
	yL.shr65bitAndLoad(L_ + 4);
	yH.shr65bitAndLoad(H_ + 4);
	V256 zL(yL), zH(yH);
	V256 xL(L_), xH(H_);
	ADD_VEC(xL, xH, yL, yH);
	zL.shlBit<12>();
	zH.shlBit<12>();
	V128 aL((L_[11] >> 22) | ((L_[12] & 3) << 10)); // [374..385]
	V128 aH((H_[11] >> 22) | ((H_[12] & 3) << 10));
	zL.x0_ |= aL;
	zH.x0_ |= aH;
	ADD_VEC(xL, xH, zH, zL); // sub
	aL.shlBit<12>();
	aH.shlBit<12>();
	ADD_VEC(xL.x0_, xH.x0_, aL, aH);
	xL.x1_ &= V128(mask);
	xH.x1_ &= V128(mask);
	xL.store(ret.L_);
	xH.store(ret.H_);
}

#if 0
template<>
template<int bit>
void F3baseDbl<F3_193>::shl()
{
	/* ignore 384, 385bit of this */
	assert(2 <= bit && bit < 32);

	V128 L2(L_ + 8), H2(H_ + 8);
	psrlq<64 - bit>(psrldq<8>(L2)).store(L_ + 12);
	psrlq<64 - bit>(psrldq<8>(H2)).store(H_ + 12);
	V128 L1(L_ + 4), H1(H_ + 4);
	(psrlq<64 - bit>(palignr<8>(L2, L1)) | psllq<bit>(L2)).store(L_ + 8);
	(psrlq<64 - bit>(palignr<8>(H2, H1)) | psllq<bit>(H2)).store(H_ + 8);

	V128 L0(L_), H0(H_);
	(psrlq<64 - bit>(palignr<8>(L1, L0)) | psllq<bit>(L1)).store(L_ + 4);
	(psrlq<64 - bit>(palignr<8>(H1, H0)) | psllq<bit>(H1)).store(H_ + 4);
	(psllq<bit>(L0) | psrlq<64 - bit>(pslldq<8>(L0))).store(L_);
	(psllq<bit>(H0) | psrlq<64 - bit>(pslldq<8>(H0))).store(H_);
}
#endif

#if 1
template<int bit, int w, int pos>
static MIE_FORCE_INLINE V128 getTri(const uint32_t *p)
{
	const int n = w == 8 ? 0 : w == 16 ? 1 : 2;
#ifdef MIE_NOT_PACK_IDX
	MIE_ALIGN(16) static const uint32_t maskTbl[3][5][4] = {
		{
			{ 0x00000000, 0x00000000, 0x00000000, 0x00000000 },
			{ 0x01010101, 0x01010101, 0x01010101, 0x01010101 },
			{ 0x03030303, 0x03030303, 0x03030303, 0x03030303 },
			{ 0x07070707, 0x07070707, 0x07070707, 0x07070707 },
			{ 0x0F0F0F0F, 0x0F0F0F0F, 0x0F0F0F0F, 0x0F0F0F0F },
		},
	};
	assert(w == 8);
	const V128& mask(maskTbl[n][bit]);
#else
	static const MIE_ALIGN(16) uint32_t maskTbl[3][2][4] = {
		{ // w = 8
			{ 0x01010101, 0x01010101, 0x01010101, 0x01010101 },
			{ 0x03030303, 0x03030303, 0x03030303, 0x03030303 },
		},
		{ // w = 16
			{ 0x00010001, 0x00010001, 0x00010001, 0x00010001 },
			{ 0x00030003, 0x00030003, 0x00030003, 0x00030003 },
		},
		{ // w = 32
			{ 0x00000001, 0x00000001, 0x00000001, 0x00000001 },
			{ 0x00000003, 0x00000003, 0x00000003, 0x00000003 },
		}
	};
	const V128& mask0(maskTbl[n][0]), mask1(maskTbl[n][1]);
#endif

	V128 t(p);
	if (pos > 0) {
		t = psrlq<pos>(t);
	}
#ifdef MIE_NOT_PACK_IDX
	return t & mask;
#else
	switch (bit) {
	case 1:
		{
			return t & mask0;
		}
	case 2:
		{
			// x = [a:b] = 2a + b
			return paddd(psrlq<1>(t) & mask0, t & mask1); // 2a + b + a
		}
		break;
	case 3:
		{
			// x = [a:b:c] = 4a + 2b + c
			V128 t1 = psrlq<2>(t) & mask0; // a
			V128 x = paddd(t1, psllq<3>(t1)); // 9a
			x = paddd(x, (t & mask1)); // 9a + 2b + c
			x = paddd(x, psrlq<1>(t) & mask0); // 9a + 3b + c
			return x;
		}
		break;
	case 4:
		/*
			x = 8a + 4b + 2c + d
			y = 27a + 9b + 3c + d = (3a + b) * 8 + (3a + b) + (2c + d) + c
		*/
		{
			V128 t1 = psrlq<2>(t) & mask1; // 2a + b
			V128 x = paddd(t1, psrlq<1>(t1) & mask0); // 2a + b + a = 3a + b
			x = paddd(psllq<3>(x), x); // 8(3a + b) + 3a + b = 27a + 9b
			x = paddd(x, t & mask1); // 27a + 9b + 2c + d
			x = paddd(x, psrlq<1>(t) & mask0); // 27a + 9b + 3c + d
			return x;
		}
		break;
	default:
		assert(0);
	}
	MIE_NOT_HERE;
#endif
}

/* T = 7 */
template<int firstBit, int bit, int w, int pos, class T>
static MIE_FORCE_INLINE void setIdx(uint8_t *p, const T& x)
{
#ifdef MIE_NOT_PACK_IDX
#if 0
	const int n = w == 8 ? 0 : w == 16 ? 1 : 2;
	MIE_ALIGN(16) static const uint32_t maskTbl[3][5][4] = {
		{
			{ 0x00000000, 0x00000000, 0x00000000, 0x00000000 },
			{ 0x01010101, 0x01010101, 0x01010101, 0x01010101 },
			{ 0x03030303, 0x03030303, 0x03030303, 0x03030303 },
			{ 0x07070707, 0x07070707, 0x07070707, 0x07070707 },
			{ 0x0F0F0F0F, 0x0F0F0F0F, 0x0F0F0F0F, 0x0F0F0F0F },
		},
	};
	assert(w == 8);
	const V128& mask(maskTbl[n][bit]);
#endif
	V128 L, H;
	L = getTri<firstBit, w, pos>(x.L_);
	H = getTri<firstBit, w, pos>(x.H_);

	(psllq<bit>(H) | L).store((uint32_t*)p);

	L = getTri<firstBit, w, pos>(x.L_ + 4);
	H = getTri<firstBit, w, pos>(x.H_ + 4);
	(psllq<bit>(H) | L).store((uint32_t*)p + 4);
#else
	V128 L, H;
	L = getTri<firstBit, w, pos>(x.L_);
	H = getTri<firstBit, w, pos>(x.H_);

	paddd(L, paddd(H, H)).store((uint32_t*)p);

	L = getTri<firstBit, w, pos>(x.L_ + 4);
	H = getTri<firstBit, w, pos>(x.H_ + 4);
	paddd(L, paddd(H, H)).store((uint32_t*)p + 4);
#endif
}

template<int bit, int w>
struct Param4makeIdxTbl {
	uint8_t (*idxTbl_)[((F3_193::N - 1 + 3) & ~3) * 4];
	const F3_193& b_;
	MIE_FORCE_INLINE Param4makeIdxTbl(uint8_t idxTbl[(w + bit - 1) / bit][((F3_193::N - 1 + 3) & ~3) * 4], const F3_193& b)
		: idxTbl_(idxTbl)
		, b_(b)
	{
	}
	template<int i>
	MIE_FORCE_INLINE void calc()
	{
		const int topBit = ((w - 1) / bit) * bit;
		const int pos = topBit - (i + 1) * bit;
		setIdx<bit, bit, w, pos>(&idxTbl_[i + 1][0], b_);
	}
private:
	void operator=(const Param4makeIdxTbl&);
};

template<int bit, int w>
static inline void makeIdxTbl(uint8_t idxTbl[(w + bit - 1) / bit][((F3_193::N - 1 + 3) & ~3) * 4], const F3_193& b)
{
	/*
		for w = 8, bit = 3
		setIdx<2, 3, 8, 6>(&idxTbl[0][0], b);
		setIdx<3, 3, 8, 3>(&idxTbl[1][0], b);
		setIdx<3, 3, 8, 0>(&idxTbl[2][0], b);
	*/
	const int topBit = ((w - 1) / bit) * bit;
	const int firstShiftBit = w - topBit < bit ? w - topBit : bit;
	setIdx<firstShiftBit, bit, w, topBit>(&idxTbl[0][0], b);

	Param4makeIdxTbl<bit, w> p(idxTbl, b);
	typedef Loop<Param4makeIdxTbl<bit, w>, 0, topBit / bit> Lp;
	Lp run(p);
#if 0
	for (int s = 1, pos = topBit - bit; pos >= 0; pos -= bit, s++) {
		setIdx<bit, bit, w>(&idxTbl[s][0], b, pos);
	}
#endif
}

template<>
template<int n>
void F3baseDbl<F3_193>::setTop(const F3_193& rhs)
{
	const int bit = (n & 31) + 32;
	V128 z(Zero());
	z.store(L_); z.store(L_ + 12);
	z.store(H_); z.store(H_ + 12);
	V128 x0(rhs.L_), x1(rhs.L_ + 4);
	(psrlq<64 - bit>(palignr<8>(x1, x0)) | psllq<bit>(x1)).store(L_ + 8);
	(psllq<bit>(x0) | psrlq<64 - bit>(pslldq<8>(x0))).store(L_ + 4);

	x0.load(rhs.H_); x1.load(rhs.H_ + 4);
	(psrlq<64 - bit>(palignr<8>(x1, x0)) | psllq<bit>(x1)).store(H_ + 8);
	(psllq<bit>(x0) | psrlq<64 - bit>(pslldq<8>(x0))).store(H_ + 4);
}
#endif

template<>
template<int bit, int w>
void F3_193::mulLatter(F3_193& out, const F3_193 *xTbl, const F3_193& b)
{
	const int loop = (w + bit - 1) / bit;
	MIE_ALIGN(16) uint8_t idxTbl[loop][((N - 1 + 3) & ~3) * 4];
	makeIdxTbl<bit, w>(idxTbl, b);
	const int topBit = ((w - 1) / bit) * bit;
	const int firstShiftBit = w - topBit < bit ? w - topBit : bit;
	F3baseDbl<F3_193> tmp;
#ifdef MIE_NOT_PACK_IDX
	{
		int idx = b.get(m - 1);
		tmp.setTop<(N - 1) * 32 - w + firstShiftBit>(xTbl[idx == 2 ? (1 << bit) : idx]);
	}
#else
	tmp.setTop<(N - 1) * 32 - w + firstShiftBit>(xTbl[b.get(m - 1)]);
#endif
#define TYPE 2

#if (TYPE == 2)
	uint32_t *L = tmp.getL();
	uint32_t *H = tmp.getH();
	V128 x0L(L), x1L(L + 4), x2L(L + 8);
	V128 x0H(H), x1H(H + 4), x2H(H + 8);
	for (int i = 0; i < loop; i++) {
		for (int j = 0; j < (N - 1) * (32 / w); j++) {
			int k = j * (w / 8);
			const F3_193& t = xTbl[idxTbl[i][k]];
			V128 tL(t.L_), tH(t.H_);
			ADD_VEC(x0L, x0H, tL, tH);
			tL.load(t.L_ + 4);
			tH.load(t.H_ + 4);
			ADD_VEC(x1L, x1H, tL, tH);
			rotrByte<w / 8>(x0L, x1L, x2L);
			rotrByte<w / 8>(x0H, x1H, x2H);
		}
		/*
			rotl<(N - 1) * 4>(x0, x1, x2) = rotl<(N - 1) * 4 - 16>(x2, x0, x1)
			N = 7
		*/
		if (i < loop - 1) {
#if 1
			if (i < loop - 2) {
				rotAndShlBitNotRet<bit>(x2L, x0L, x1L);
				rotAndShlBitNotRet<bit>(x2H, x0H, x1H);
			} else {
				L[12] = rotAndShlBit<bit>(x2L, x0L, x1L);
				H[12] = rotAndShlBit<bit>(x2H, x0H, x1H);
			}
#else
			rotlByte<(N - 1) * 4 - 16>(x2L, x0L, x1L);
			rotlByte<(N - 1) * 4 - 16>(x2H, x0H, x1H);
			if (i < loop - 2) {
				shlBitNotRet<bit>(x2L, x0L, x1L);
				shlBitNotRet<bit>(x2H, x0H, x1H);
			} else {
				L[12] = shlBit<bit>(x2L, x0L, x1L);
				H[12] = shlBit<bit>(x2H, x0H, x1H);
			}
#endif
			V128 t;
			t = x0L; x0L = x2L; x2L = x1L; x1L = t;
			t = x0H; x0H = x2H; x2H = x1H; x1H = t;
		} else {
			rotlByte<(N - 1) * 4 - 16>(x2L, x0L, x1L);
			rotlByte<(N - 1) * 4 - 16>(x2H, x0H, x1H);
			x2L.store(L); x0L.store(L + 4); x1L.store(L + 8);
			x2H.store(H); x0H.store(H + 4); x1H.store(H + 8);
		}
	}
#endif

#if (TYPE == 1)
	for (int i = 0; i < loop; i++) {
		uint32_t *L = tmp.getL();
		uint32_t *H = tmp.getH();
		V128 x0L(L), x1L(L + 4), x2L(L + 8);
		V128 x0H(H), x1H(H + 4), x2H(H + 8);
		for (int j = 0; j < (N - 1) * (32 / w); j++) {
			int k = j * (w / 8);
			const F3_193& t = xTbl[idxTbl[i][k]];
			V128 tL(t.L_), tH(t.H_);
			ADD_VEC(x0L, x0H, tL, tH);
			tL.load(t.L_ + 4);
			tH.load(t.H_ + 4);
			ADD_VEC(x1L, x1H, tL, tH);
			rotrByte<w / 8>(x0L, x1L, x2L);
			rotrByte<w / 8>(x0H, x1H, x2H);
		}
		/*
			rotl<(N - 1) * 4>(x0, x1, x2) = rotl<(N - 1) * 4 - 16>(x2, x0, x1)
		*/
		rotlByte<(N - 1) * 4 - 16>(x2L, x0L, x1L);
		rotlByte<(N - 1) * 4 - 16>(x2H, x0H, x1H);
		x2L.store(L); x0L.store(L + 4); x1L.store(L + 8);
		x2H.store(H); x0H.store(H + 4); x1H.store(H + 8);
		if (i < loop - 1) tmp.shl<bit>();
	}
#endif
#if (TYPE == 0)
	for (int i = 0; i < loop; i++) {
		for (int j = 0; j < (N - 1) * (32 / w); j++) {
			int k = j * (w / 8);
			tmp.add(xTbl[idxTbl[i][k]]);
			tmp.rotrByte<w / 8>();
		}
		tmp.rotlByte<(N - 1) * 4>();
		if (i < loop - 1) tmp.shl<bit>();
	}
#endif
	tmp.reduce(out); // 353clk
}

#if 1
template<int shift>
static MIE_FORCE_INLINE void expand64(uint32_t *out, uint64_t x, uint32_t bit192 = 0)
{
	out[6] = bit192;
	out[7] = 0;
	expand32<shift>(out, (uint32_t)x);
	expand32<shift>(out + 3, (uint32_t)(x >> 32));
}
#endif

template<int shift>
static MIE_FORCE_INLINE void expand64(uint32_t *out, uint32_t xL, uint32_t xH, uint32_t bit192 = 0)
{
	const uint32_t m10 = GetMask(10);
	const uint32_t m11 = GetMask(11);

	switch (shift) {
	case 0:
		out[0] = cubeTbl[xL & m11];
		out[1] = cubeTbl[(xL >> 11) & m11] << 1;
		out[2] = cubeTbl[xL >> 22] << 2;
		out[3] = cubeTbl[xH & m11];
		out[4] = cubeTbl[(xH >> 11) & m11] << 1;
		out[5] = cubeTbl[xH >> 22] << 2;
		out[6] = bit192;
		out[7] = 0;
		return;
	case 1:
		out[0] = cubeTbl[xL & m11] << 1;
		out[1] = cubeTbl[(xL >> 11) & m10] << 2;
		out[2] = cubeTbl[xL >> 21];
		out[3] = cubeTbl[xH & m11] << 1;
		out[4] = cubeTbl[(xH >> 11) & m10] << 2;
		out[5] = cubeTbl[xH >> 21];
		out[6] = bit192;
		out[7] = 0;
		return;
	case 2:
		out[0] = cubeTbl[xL & m10] << 2;
		out[1] = cubeTbl[(xL >> 10) & m11];
		out[2] = cubeTbl[xL >> 21] << 1;
		out[3] = cubeTbl[xH & m10] << 2;
		out[4] = cubeTbl[(xH >> 10) & m11];
		out[5] = cubeTbl[xH >> 21] << 1;
		out[6] = bit192;
		out[7] = 0;
		return;
	}
	MIE_NOT_HERE;
}

template<>
void F3_193::cube(F3base& x)
{
#if 1
#if 1
	/* 164clk */
	F3_193 a0, a1, a2;

	{
		uint32_t xL = x.L_[0]; // [0..31]
		uint32_t xH = x.H_[0];
		uint32_t tL = (x.L_[5] >> 25) | ((x.L_[6] & 1) << 7); // [185..192]
		uint32_t tH = (x.H_[5] >> 25) | ((x.H_[6] & 1) << 7);
		uint32_t yL = (tL >> 4) | (tH << 4); // [189..192] - ([185..192] << 4)
		uint32_t yH = (tH >> 4) | (tL << 4);
		rawAdd(&xL, &xH, yL, yH);
		yL = tL & GetMask(4); // [185..188]
		yH = tH & GetMask(4);
		rawAdd(&xL, &xH, yL, yH);
		expand64<0>(a0.L_, xL, x.L_[1], x.L_[2] & 1);
		expand64<0>(a0.H_, xH, x.H_[1], x.H_[2] & 1);
	}

	{
		V128 xL = psrlq<1>(V128(x.L_[4]) | pslldq<4>(x.L_[5])) | psllq<63>(x.L_[6]); // [129..192]
		V128 xH = psrlq<1>(V128(x.H_[4]) | pslldq<4>(x.H_[5])) | psllq<63>(x.H_[6]);
		V128 yL = xL;
		V128 yH = xH;
		V128 t1L = psllq<4>(xL), t1H = psllq<4>(xH);
		ADD_VEC(yL, yH, t1L, t1H); // [129..188] << 4
		uint32_t tL = (x.L_[3] >> 29) | ((x.L_[4] & 1) << 3); // [125..128]
		uint32_t tH = (x.H_[3] >> 29) | ((x.H_[4] & 1) << 3);
		t1L = pslldq<1>(xL) | V128(tH | (tL << 4));
		t1H = pslldq<1>(xH) | V128(tL | (tH << 4));
		ADD_VEC(yL, yH, t1L, t1H);
		expand64<1>(a1.L_, yL.get32bit(), psrldq<4>(yL).get32bit());
		expand64<1>(a1.H_, yH.get32bit(), psrldq<4>(yH).get32bit());
	}

	{
		V128 xL = V128(x.L_[2]) | pslldq<4>(x.L_[3]); // [64..127]
		V128 xH = V128(x.H_[2]) | pslldq<4>(x.H_[3]);
		xL = psrlq<1>(xL) | psllq<63>(V128(x.L_[4])); // [65..128]
		xH = psrlq<1>(xH) | psllq<63>(V128(x.H_[4]));
		V128 yL = psllq<4>(xL); // [65..124] << 4
		V128 yH = psllq<4>(xH);
		ADD_VEC(xL, xH, yH, yL); // sub(xL, xH, yL, yH);

		expand64<2>(a2.L_, xL.get32bit(), psrldq<4>(xL).get32bit());
		expand64<2>(a2.H_, xH.get32bit(), psrldq<4>(xH).get32bit());
	}

	x = a0 + a1 + a2;
#else
	F3_193 a0, a1, a2;

	{
		uint64_t xL = x.L_[0] | (uint64_t(x.L_[1]) << 32); // [0..63]
		uint64_t xH = x.H_[0] | (uint64_t(x.H_[1]) << 32);
		uint32_t tL = (x.L_[5] >> 25) | ((x.L_[6] & 1) << 7); // [185..192]
		uint32_t tH = (x.H_[5] >> 25) | ((x.H_[6] & 1) << 7);
		uint64_t yL = (tL >> 4) | (tH << 4); // [189..192] - ([185..192] << 4)
		uint64_t yH = (tH >> 4) | (tL << 4);
		rawAdd(&xL, &xH, yL, yH);
		yL = tL & GetMask(4); // [185..188]
		yH = tH & GetMask(4);
		rawAdd(&xL, &xH, yL, yH);
		expand64<0>(a0.L_, xL, x.L_[2] & 1);
		expand64<0>(a0.H_, xH, x.H_[2] & 1);
	}

	{
		uint64_t xL = ((x.L_[4] | (uint64_t(x.L_[5]) << 32)) >> 1) | (uint64_t(x.L_[6] & 1) << 63); // [129..192]
		uint64_t xH = ((x.H_[4] | (uint64_t(x.H_[5]) << 32)) >> 1) | (uint64_t(x.H_[6] & 1) << 63);
		uint64_t yL = xL;
		uint64_t yH = xH;
		rawAdd(&yL, &yH, xL << 4, xH << 4); // [129..188] << 4
		uint32_t tL = (x.L_[3] >> 29) | ((x.L_[4] & 1) << 3); // [125..128]
		uint32_t tH = (x.H_[3] >> 29) | ((x.H_[4] & 1) << 3);
		rawAdd(&yL, &yH, (xL << 8) | tH | (tL << 4), (xH << 8) | tL | (tH << 4));
		expand64<1>(a1.L_, yL);
		expand64<1>(a1.H_, yH);
	}

	{
		uint64_t xL = x.L_[2] | (uint64_t(x.L_[3]) << 32); // [64..127]
		uint64_t xH = x.H_[2] | (uint64_t(x.H_[3]) << 32);
		xL >>= 1; xL |= uint64_t(x.L_[4]) << 63; // [65..128]
		xH >>= 1; xH |= uint64_t(x.H_[4]) << 63;
		uint64_t yL = xL << 4; // [65..124] << 4
		uint64_t yH = xH << 4;
		rawSub(&xL, &xH, yL, yH);

		expand64<2>(a2.L_, xL);
		expand64<2>(a2.H_, xH);
	}

	x = a0 + a1 + a2;
#endif
#else
	uint32_t L[N * 3];
	uint32_t H[N * 3];
	rawClear(L);
	rawClear(H);
	for (int i = 0; i < m; i++) {
		rawSet(L, H, i * 3, x.get(i));
	}
	for (int i = m * 3 - 1; i >= m; i--) {
		uint32_t t, x = rawGet(L, H, i);
		t = (rawGet(L, H, i - (m - ma)) - x + 3) % 3;
		rawSet(L, H, i - (m - ma), t);
		t = (rawGet(L, H, i - m) + x) % 3;
		rawSet(L, H, i - m, t);
		rawSet(L, H, i, 0);
	}
	x = F3_193(L, H);
#endif
}

template<>
void F3_193::rawShl(F3base& x, int bit)
{
	V128 L0(x.L_), H0(x.H_);
	if (bit >= 128) {
		bit -= 128;
		if (bit > 0) {
			V128 shift(bit);
			V128 revShift(64 - bit);
			(psllq(L0, shift) | psrlq(pslldq<8>(L0), revShift)).store(x.L_ + 4);
			(psllq(H0, shift) | psrlq(pslldq<8>(H0), revShift)).store(x.H_ + 4);
		} else {
			L0.store(x.L_ + 4);
			H0.store(x.H_ + 4);
		}
		V128 z = Zero();
		z.store(x.L_);
		z.store(x.H_);
		return;
	}
	/*
	[x1H:x1L] [x0H:x0L] << bit
	= ([x1L:x0H] [x0L:0]) << (64 - bit)
	= [*:(x1L:x0H) << (bit - 64)] | ([x0L:0] << (bit - 64))
	*/
	V128 L1(x.L_ + 4), H1(x.H_ + 4);
	if (bit == 64) {
		palignr<8>(L1, L0).store(x.L_ + 4);
		palignr<8>(H1, H0).store(x.H_ + 4);
		pslldq<8>(L0).store(x.L_);
		pslldq<8>(H0).store(x.H_);
		return;
	}
	if (bit >= 64) {
		V128 shift(bit - 64);
		V128 revShift(128 - bit);
		(psllq(palignr<8>(L1, L0), shift) | psrlq(L0, revShift)).store(x.L_ + 4);
		(psllq(palignr<8>(H1, H0), shift) | psrlq(H0, revShift)).store(x.H_ + 4);
		psllq(pslldq<8>(L0), shift).store(x.L_);
		psllq(pslldq<8>(H0), shift).store(x.H_);
		return;
	}
	V128 shift(bit);
	V128 revShift(64 - bit);
	// [x1L:x0H] >> (64 - bit) | (x1L << bit)
	(psrlq(palignr<8>(L1, L0), revShift) | psllq(L1, shift)).store(x.L_ + 4);
	(psrlq(palignr<8>(H1, H0), revShift) | psllq(H1, shift)).store(x.H_ + 4);
	(psllq(L0, shift) | psrlq(pslldq<8>(L0), revShift)).store(x.L_);
	(psllq(H0, shift) | psrlq(pslldq<8>(H0), revShift)).store(x.H_);
}

/*
	./mk_const.sh "m=193;(3^m-1)/4+1"
*/
template<>
const uint32_t F3_193::powTblForSquareRoot[] = {
	0x02116941,0x1D373FE2,0xFDA0628D,0xE7B9E1B5,0x8DEFFA16,0x231DC649,0xF134C2E0,0x0648DEB6,0x7EC73096,0xEE7C,
};
template<>
const uint32_t F3_193::powTblSize = mie::NumOfArray(F3_193::powTblForSquareRoot);

#endif /* MIE_MIE_SRC_F3BASE193_H_ */
