#ifndef MIE_ETAT_SRC_F3BASE97_H_
#define MIE_ETAT_SRC_F3BASE97_H_

typedef MIE_ETAT::F3base<97, 12> F3_97;

#include "etat_impl.h"


static OptimizedMulLatter optimizedMulLatter;
static OptimizedMakePrevMulTbl optimizedMakePrevMulTbl;
static OptimizedCube optimizedCube;
static OptimizedRawShl optimizedRawShl;
static OptimizedDegree optimizedDegree;
static OptimizedCubeSeq optimizedCubeSeq;

template<>
template<int bit, int w>
inline void F3_97::mulLatter(F3base& out, const F3base *xTbl, const F3base& b)
{
	((F3_97::mulLatterFunc)optimizedMulLatter.get<bit, w>())(out, xTbl, b);
}
template<>
template<int bit>
inline void F3_97::makePrevMulTbl(F3base *xTbl, const F3base& x)
{
	optimizedMakePrevMulTbl.get<bit>()(xTbl, x);
}

template<>
inline void F3_97::cube(F3base& x)
{
	optimizedCube.get()(x);
}

template<>
inline void F3_97::cubeSeq(F3base *tbl)
{
	optimizedCubeSeq.get()(tbl);
}

template<>
inline void F3_97::rawShl(F3base& x, int bit)
{
	optimizedRawShl.get()(x, bit);
}
template<>
inline int F3_97::degree(const F3base& x)
{
	return optimizedDegree.get()(x);
}

/*
	./mk_const.sh "m=97;(3^m-1)/4+1"
*/
template<>
const uint32_t F3_97::powTblForSquareRoot[] = {
	0x76111CA1, 0x2A487C6B, 0xC337211B, 0xFC370813, 0x00D5FC08,
};
template<>
const uint32_t F3_97::powTblSize = mie::NumOfArray(F3_97::powTblForSquareRoot);

template<>
MIE_FORCE_INLINE void F3_97::neg(F3base& out, const F3base& a)
{
	V128 t(a.L_);
	copy128(out.L_, a.H_);
	t.store(out.H_);
}

template<>
MIE_FORCE_INLINE void F3_97::add(F3base& out, const F3base& a, const F3base& b)
{
	V128 aL(a.L_), aH(a.H_), bL(b.L_), bH(b.H_);
#ifdef MIE_USE_6OP_ADD
	bL ^= aL;
	bH ^= aH;
	aL ^= bH;
	aH ^= bL;
	andn(bH, aH).store(out.L_);
	andn(bL, aL).store(out.H_);
#else
	V128 t((aL | aH) & (bL | bH));
	(t ^ (aL | bL)).store(out.L_);
	(t ^ (aH | bH)).store(out.H_);
#endif
}

template<>
MIE_FORCE_INLINE void F3_97::sub(F3base& out, const F3base& a, const F3base& b)
{
	V128 aL(a.L_), aH(a.H_), bL(b.H_), bH(b.L_);
#ifdef MIE_USE_6OP_ADD
	bL ^= aL;
	bH ^= aH;
	aL ^= bH;
	aH ^= bL;
	andn(bH, aH).store(out.L_);
	andn(bL, aL).store(out.H_);
#else
	V128 t((aL | aH) & (bL | bH));
	(t ^ (aL | bL)).store(out.L_);
	(t ^ (aH | bH)).store(out.H_);
#endif
}

template<>
MIE_FORCE_INLINE void F3_97::addsub(F3base& addOut, F3base& subOut, const F3base& a, const F3base& b)
{
	V128 bL(b.L_), bH(b.H_), tb(bH | bL), aL(a.L_), aH(a.H_), t((aL | aH) & tb);
	((aL | bL) ^ t).store(addOut.L_);
	((aL | bH) ^ t).store(subOut.L_);
	((aH | bH) ^ t).store(addOut.H_);
	((aH | bL) ^ t).store(subOut.H_);
}

template<>
MIE_FORCE_INLINE void F3_97::addPara(F3base& out1, F3base& out2, const F3base& a1, const F3base& a2, const F3base& b)
{
	V128 bL(b.L_), bH(b.H_), tb(bH | bL), aL(a1.L_), aH(a1.H_), t((aL | aH) & tb);
	((aL | bL) ^ t).store(out1.L_);
	((aH | bH) ^ t).store(out1.H_);
	aL.load(a2.L_);
	aH.load(a2.H_);
	t = (aL | aH) & tb;
	((aL | bL) ^ t).store(out2.L_);
	((aH | bH) ^ t).store(out2.H_);
}

template<>
inline bool F3_97::isZero() const
{
	return (V128(L_) | V128(H_)).cmpeq(Zero()).pmovmskb() == 0xffff;
}

template<>
inline bool F3_97::operator==(const F3base& rhs) const
{
	V128 L(L_), H(H_);
	L.cmpeq(V128(rhs.L_));
	H.cmpeq(V128(rhs.H_));
	return (L & H).pmovmskb() == 0xffff;
}

template<>
inline void F3_97::swap(F3base& rhs)
{
	swap128(rhs.L_, L_);
	swap128(rhs.H_, H_);
}

template<>
inline void inverse(F3_97& x)
{
	/*
		cube𑬂ȂƂ炪ȂȂ
		cubeȂ̂łI
		cube : 1 + 2 + 4 + 8 + 16 + 32 + 32 + 1 = 96
		mul  : 9
	*/
	F3_97 y(x), z(x), w;
	F3_97::cube(z); // (10)_3
	y *= z; // (11)_3
	z = y;

	F3_97::cubeN(z, 2); // (1100)_3
	y *= z; // (1111)_3
	z = y;

	F3_97::cubeN(z, 4); // (11110000)_3
	y *= z; // (11111111)_3
	z = y;

	F3_97::cubeN(z, 8); // (1111111100000000)_3
	y *= z; // (1x16)_3
	z = y;

	F3_97::cubeN(z, 16); // (1x16 0x16)_3
	y *= z; // (1x32)_3
	w = y;

	F3_97::cubeN(w, 32); // (1x32 0x32)_3
#if 1
	F3_97 xTbl[intpow3T<g_pbit>::val];
	F3_97::makePrevMulTbl<g_pbit>(xTbl, y);
	F3_97::mulLatter<g_pbit, g_w>(w, xTbl, w);
	F3_97::cubeN(w, 32); // (1x64 0x32)_3
	F3_97::mulLatter<g_pbit, g_w>(w, xTbl, w);
#else
	w *= y; // (1x64)_3

	F3_97::cubeN(w, 32); // (1x64 0x32)_3
	w *= y; // (1x96)_3
#endif

	w *= w; // (2x96)_3
	F3_97::cube(w); // (2x96 0)_3
	x *= w; // (2x96 1)_3
}

#if 0
template<>
template<int bit>
void F3_97::mulLatterC(F3_97& out, const F3_97 *xTbl, const F3_97& b)
{
	const int loop = (g_w + bit - 1) / bit;
	MIE_ALIGN(16) uint8_t idxTbl[loop][((N - 1 + 3) & ~3) * 4];
	makeIdxTbl<bit, g_w>(idxTbl, b);
	const int topBit = ((g_w - 1) / bit) * bit;
	const int firstShiftBit = g_w - topBit < bit ? g_w - topBit : bit;
	F3baseDbl<F3_97> tmp;
	tmp.setTop<(N - 1) * 32 - g_w + firstShiftBit>(xTbl[b.get(m - 1)]);
	for (int i = 0; i < loop; i++) {
		for (int j = 0; j < (N - 1) * (32 / g_w); j++) {
			int k = j * (g_w / 8);
			tmp.add(xTbl[idxTbl[i][k]], k);
		}
		if (i < loop - 1) tmp.shl<bit>();
	}
	out = tmp.reduce(); // 353clk
}
#endif

#endif /* MIE_ETAT_SRC_F3BASE97_H_ */
