#ifndef MIE_ETAT_M193_IF_H_
#define MIE_ETAT_M193_IF_H_
/**
	interface for m = 193
*/
namespace m193 {

struct MIE_Fp *FpCreateByStr(const char *str)
{
	F3_193 *a = (F3_193*)AlignedMalloc(sizeof(F3_193));
	if (a == 0) return 0;
	a->setByString(str);
	return (MIE_Fp*)a;
}

struct MIE_Fp *FpCreate(void)
{
	return FpCreateByStr(0);
}

void FpDestroy(MIE_Fp *in)
{
	AlignedFree(in);
}

template<typename Abst, typename Conc>
MIE_Err toStr(char *buf, size_t bufSize, const Abst *in)
{
	try {
		Conc *a = (Conc*)in;
		std::string s = a->toString();
		if (s.size() + 1 <= bufSize) {
			strcpy(buf, s.c_str());
			return MIE_NO_ERR;
		}
	} catch (...) {
	}
	return MIE_SMALL_SIZE;
}

// Fp

MIE_Err FpToStr(char *buf, size_t bufSize, const MIE_Fp *in)
{
	return toStr<MIE_Fp, F3_193>(buf, bufSize, in);
}

int FpEqual(const MIE_Fp *in1, const MIE_Fp *in2)
{
	return *(const F3_193*)in1 == *(const F3_193*)in2;
}

void FpClear(MIE_Fp *in)
{
	((F3_193*)in)->clear();
}

void FpSet(MIE_Fp *out, const MIE_Fp *in)
{
	if (out == in) return;
	*(F3_193*)out = *(const F3_193*)in;
}

void FpSetByStr(MIE_Fp *out, const char *str)
{
	((F3_193*)out)->setByString(str);
}

void FpAdd(MIE_Fp *out, const MIE_Fp *in1, const MIE_Fp *in2)
{
	F3_193::add(*(F3_193*)out, *(const F3_193*)in1, *(const F3_193*)in2);
}

void FpSub(MIE_Fp *out, const MIE_Fp *in1, const MIE_Fp *in2)
{
	F3_193::sub(*(F3_193*)out, *(const F3_193*)in1, *(const F3_193*)in2);
}

void FpMul(MIE_Fp *out, const MIE_Fp *in1, const MIE_Fp *in2)
{
	F3_193::mul(*(F3_193*)out, *(const F3_193*)in1, *(const F3_193*)in2);
}

void FpNeg(MIE_Fp *out, const MIE_Fp *in)
{
	F3_193::neg(*(F3_193*)out, *(const F3_193*)in);
}

void FpInv(MIE_Fp *out, const MIE_Fp *in)
{
	FpSet(out, in);
	inverse(*(F3_193*)out);
}

int FpSqrt(MIE_Fp *out, const MIE_Fp *in)
{
	FpSet(out, in);
	return F3_193::squareRoot(*(F3_193*)out);
}

void FpPow(MIE_Fp *out, const MIE_Fp *in, const unsigned int *tbl, size_t tblSize)
{
	FpSet(out, in);
	power(*(F3_193*)out, tbl, tblSize);
}

// Fq

MIE_Fq *FqCreate(void)
{
	F3_193_6 *a = (F3_193_6*)AlignedMalloc(sizeof(F3_193_6));
	if (a == 0) return 0;
	a->clear();
	return (MIE_Fq*)a;
}

void FqDestroy(MIE_Fq *in)
{
	AlignedFree(in);
}

MIE_Err FqToStr(char *buf, size_t bufSize, const MIE_Fq *in)
{
	return toStr<MIE_Fq, F3_193_6>(buf, bufSize, in);
}

int FqEqual(const MIE_Fq *in1, const MIE_Fq *in2)
{
	return *(const F3_193_6*)in1 == *(const F3_193_6*)in2;
}

void FqClear(MIE_Fq *in)
{
	((F3_193_6*)in)->clear();
}

void FqSet(MIE_Fq *out, const MIE_Fq *in)
{
	if (out == in) return;
	*(F3_193_6*)out = *(const F3_193_6*)in;
}

MIE_Fp *FqGetRef(MIE_Fq *out, size_t idx)
{
	if (idx >= 6) return 0;
	return (MIE_Fp*)&((F3_193*)out)[idx];
}

const MIE_Fp *FqGetConstRef(const MIE_Fq *out, size_t idx)
{
	if (idx >= 6) return 0;
	return (const MIE_Fp*)&((const F3_193*)out)[idx];
}

void FqSetFp(MIE_Fq *out, size_t idx, const MIE_Fp *in)
{
	MIE_Fp *p = FqGetRef(out, idx);
	if (p) FpSet(p, in);
}

void FqAdd(MIE_Fq *out, const MIE_Fq *in1, const MIE_Fq *in2)
{
	F3_193_6 t(*(const F3_193_6*)in1);
	t += *(const F3_193_6*)in2;
	*(F3_193_6*)out = t;
}

void FqSub(MIE_Fq *out, const MIE_Fq *in1, const MIE_Fq *in2)
{
	F3_193_6 t(*(const F3_193_6*)in1);
	t -= *(const F3_193_6*)in2;
	*(F3_193_6*)out = t;
}

void FqMul(MIE_Fq *out, const MIE_Fq *in1, const MIE_Fq *in2)
{
	F3_193_6 t(*(const F3_193_6*)in1);
	t *= *(const F3_193_6*)in2;
	*(F3_193_6*)out = t;
}

void FqPow(MIE_Fq *out, const MIE_Fq *in, const unsigned int *tbl, size_t tblSize)
{
	FqSet(out, in);
	power(*(F3_193_6*)out, tbl, tblSize);
}

// functions for Ec
MIE_Ec *EcCreate(void)
{
	EConF3_193 *a = (EConF3_193*)AlignedMalloc(sizeof(EConF3_193));
	if (a == 0) return 0;
	a->x.clear();
	a->y.clear();
	return (MIE_Ec*)a;
}
void EcDestroy(MIE_Ec *in)
{
	AlignedFree(in);
}

MIE_Err EcToStr(char *buf, size_t bufSize, const MIE_Ec *in)
{
	return toStr<MIE_Ec, EConF3_193>(buf, bufSize, in);
}

int EcSetX(MIE_Ec *out, const MIE_Fp *xCandidate)
{
	return ((EConF3_193*)out)->set(*(const F3_193*)xCandidate, true);
}

void EcSet(MIE_Ec *out, const MIE_Ec *in)
{
	if (out == in) return;
	*(EConF3_193*)out = *(const EConF3_193*)in;
}

MIE_Fp *EcGetRefX(MIE_Ec *in)
{
	return (MIE_Fp*)&((EConF3_193*)in)->x;
}

MIE_Fp *EcGetRefY(MIE_Ec *in)
{
	return (MIE_Fp*)&((EConF3_193*)in)->y;
}

const MIE_Fp *EcGetConstRefX(const MIE_Ec *in)
{
	return (const MIE_Fp*)&((const EConF3_193*)in)->x;
}

const MIE_Fp *EcGetConstRefY(const MIE_Ec *in)
{
	return (const MIE_Fp*)&((const EConF3_193*)in)->y;
}

int EcValid(const MIE_Ec *in)
{
	return ((const EConF3_193*)in)->isValid();
}

void EcAdd(MIE_Ec *out, const MIE_Ec *in1, const MIE_Ec *in2)
{
	EConF3_193 t(*(const EConF3_193*)in1);
	t += *(const EConF3_193*)in2;
	*(EConF3_193*)out = t;
}

void EcNeg(MIE_Ec *out, const MIE_Ec *in)
{
	EcSet(out, in);
	((EConF3_193*)out)->neg();
}

int EcEqual(const MIE_Ec *in1, const MIE_Ec *in2)
{
	return *(const EConF3_193*)in1 == *(const EConF3_193*)in2;
}

void EcPow(MIE_Ec *out, const MIE_Ec *in, const unsigned int *tbl, size_t tblSize)
{
	EcSet(out, in);
	EConF3_193& x(*(EConF3_193*)out);
	EConF3_193 t(x);
	x.x.clear();
	x.y.clear();
	for (size_t i = 0; i < tblSize; i++) {
		uint32_t n = tbl[i];
		for (int j = 0; j < 32; j++) {
			if (n & (1U << j)) {
				x += t;
			}
			t += t;
		}
	}
}

void pairing(MIE_Fq *out, const MIE_Ec *in1, const MIE_Ec *in2)
{
	*(F3_193_6*)out = CalcEtatPairing(*(const EConF3_193*)in1, *(const EConF3_193*)in2);
}

void destorySystem(MIE_System *)
{
	/* none to do */
}

MIE_Err createSystem(MIE_System **psys)
{
	static MIE_System sys = {
		destorySystem,
		// Fp
		FpCreate,
		FpCreateByStr,
		FpDestroy,
		FpToStr,
		FpEqual,
		FpClear,

		FpSet,
		FpSetByStr,

		FpAdd,
		FpSub,
		FpMul,
		FpInv,
		FpNeg,
		FpSqrt,
		FpPow,

		// Fq
		FqCreate,
		FqDestroy,
		FqToStr,
		FqEqual,
		FqClear,

		FqSet,
		FqGetRef,
		FqGetConstRef,
		FqSetFp,

		FqAdd,
		FqSub,
		FqMul,
		// FqInv,
		// FqNeg,
		FqPow,

		// Ec
		EcCreate,
		EcDestroy,

		EcToStr,

		EcSetX,

		EcSet,
		EcGetRefX,
		EcGetRefY,
		EcGetConstRefX,
		EcGetConstRefY,
		EcValid,

		EcAdd,
		EcNeg,
		EcEqual,
		EcPow,

		pairing,
	};
	*psys = &sys;
	return MIE_NO_ERR;
}

} // end of m193

#endif
