#ifndef MIE_ETAT_SRC_MULTI_CORE_H_
#define MIE_ETAT_SRC_MULTI_CORE_H_

#include "mie_thread.h"

namespace MIE_ETAT {

class Parallelizer : public mie::ThreadBase<Parallelizer> {
	enum State {
		WAIT,
		BUSY_LOOP,
		RUNNING,
		QUIT
	};
	mie::Event event_;
	volatile State state_;
	void *para_;
	void (*func_)(void*);
public:
	Parallelizer()
		: state_(WAIT)
		, para_(0)
		, func_(0)
	{
	}
	void threadEntry()
	{
		do {
			event_.wait();
			for (;;) {
				/* BUSY_LOOP or RUNNING */
				do { _mm_pause(); } while (state_ == BUSY_LOOP); // busy loop
				if (state_ != RUNNING) break;
				func_(para_);
				_mm_sfence();
				state_ = BUSY_LOOP;
			}
		} while (state_ != QUIT);
	}
	void stopThread()
	{
		state_ = QUIT;
	}
	void wakeup()
	{
		state_ = BUSY_LOOP;
		event_.wakeup();
	}
	void suspend()
	{
		state_ = WAIT;
	}
	void calc(void *para1, void *para2, void func(void*))
	{
		para_ = para1;
		func_ = func;
		_mm_sfence();
		state_ = RUNNING;
		func(para2);
		while (state_ == RUNNING) _mm_pause();
	}
	void init()
	{
		beginThread();
#ifdef MIE_USE_MULTI_CORE_WITH_BUSY_LOOP
		wakeup();
#endif
	}
};

#ifdef MIE_ETAT_MAIN_INSTANCE_IS_HERE_

static Parallelizer s_parallelizer;

#endif

} // end of namespace

#endif
