#include "v128.h"

struct V256 {
	V128 x0_, x1_;
	V256()
	{
	}
	V256(const uint32_t *p)
		: x0_(p)
		, x1_(p + 4)
	{
	}
	V256(const uint32_t *L, const uint32_t *H)
		: x0_(L)
		, x1_(H)
	{
	}
	MIE_FORCE_INLINE V256(const V256& rhs)
		: x0_(rhs.x0_)
		, x1_(rhs.x1_)
	{
	}
	MIE_FORCE_INLINE V256& operator|=(const V256& rhs)
	{
		x0_ |= rhs.x0_;
		x1_ |= rhs.x1_;
		return *this;
	}
	MIE_FORCE_INLINE V256& operator&=(const V256& rhs)
	{
		x0_ &= rhs.x0_;
		x1_ &= rhs.x1_;
		return *this;
	}
	MIE_FORCE_INLINE V256& operator^=(const V256& rhs)
	{
		x0_ ^= rhs.x0_;
		x1_ ^= rhs.x1_;
		return *this;
	}
	MIE_FORCE_INLINE V256& andn(const V256& rhs)
	{
		x0_.andn(rhs.x0_);
		x1_.andn(rhs.x1_);
		return *this;
	}
	void store(uint32_t *p)
	{
		x0_.store(p);
		x1_.store(p + 4);
	}
	void load(const uint32_t *p)
	{
		x0_.load(p);
		x1_.load(p + 4);
	}
	uint32_t pmovmskbL() const { return x0_.pmovmskb(); }
	uint32_t pmovmskbH() const { return x1_.pmovmskb(); }
	V256& cmpeq(const V256& rhs)
	{
		x0_.cmpeq(rhs.x0_);
		x1_.cmpeq(rhs.x1_);
		return *this;
	}
	/**
		[x1_:x0_] >>= n
	*/
	template<int n>
	void shrBit()
	{
		const int q = n / 8;
		const int r = n & 7;
		V128 t(x0_);
		if (q > 0) {
			x0_ = palignr<q>(x1_, x0_);
		}
		t = palignr<q + 1>(x1_, t);
		x0_ = psrlq<r>(x0_);
		t = psllq<8 - r>(t);
		x0_ |= t;
		x1_.shrBit<r>();
	}
	/**
		[x1_:x0_] <<= n
	*/
	template<int n>
	void shlBit()
	{
		const int q = n / 8;
		const int r = n & 7;
		if (q > 0) {
			x1_ = palignr<16 - q>(x1_, x0_);
			x0_ = pslldq<q>(x0_);
		}
		V128 t = palignr<15>(x1_, x0_);
		t = psrlq<8 - r>(t);
		x1_ = psllq<r>(x1_);
		x1_ |= t;
		t = pslldq<1>(x0_);
		x0_ = psllq<r>(x0_);
		t = psrlq<8 - r>(t);
		x0_ |= t;
	}
	/*
		p
		0 1|2 3|4 5|6
	*/
	void MIE_FORCE_INLINE shr65bitAndLoad(const uint32_t *p)
	{
		load(p);
		x0_ = palignr<8>(x1_, x0_);
		x1_ = psrldq<8>(x1_);
		V128 t(*(p + 8) & 3);
		t = pslldq<8>(t);
		x1_ |= t;
		shrBit<1>();
	}
	std::string toString() const
	{
		return x1_.toString() + ':' + x0_.toString();
	}
};

static MIE_FORCE_INLINE V256 andn(const V256& lhs, const V256& rhs)
{
	return V256(lhs).andn(rhs);
}

static MIE_FORCE_INLINE V256 operator|(const V256& lhs, const V256& rhs)
{
	return V256(lhs) |= rhs;
}

static MIE_FORCE_INLINE V256 operator&(const V256& lhs, const V256& rhs)
{
	return V256(lhs) &= rhs;
}

static MIE_FORCE_INLINE V256 operator^(const V256& lhs, const V256& rhs)
{
	return V256(lhs) ^= rhs;
}

static MIE_FORCE_INLINE void swap256(uint32_t *p, uint32_t *q)
{
	V256 t(p);
	V256(q).store(p);
	t.store(q);
}

static MIE_FORCE_INLINE void copy256(uint32_t *dest, const uint32_t *src)
{
	V256(src).store(dest);
}
