#ifndef ETAT_TEST_UTIL_H_
#define ETAT_TEST_UTIL_H_

#include <algorithm>
#include <stdio.h>

#if 1
#define clkbegin()
#define clkend()
#define clkput()
#endif

#define UASSERT(x) uassert(#x, __FILE__, __LINE__, x)

#ifdef __GNUC__
#include <sys/time.h>
#include <stdio.h>

static inline double gettimeofday_sec()
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return tv.tv_sec + (double) tv.tv_usec * 1e-6;
}
#else
#include <time.h>
static inline double gettimeofday_sec()
{
	return clock() / double(CLOCKS_PER_SEC);
}
#endif

static inline void uassert(const char *msg, const char *file, int line, bool b)
{
	if (b) return;
	printf("%s:%d error [%s]\n", file, line, msg);
}

template<int N>
static bool verifyOnlyOneBit(const uint32_t (&buf)[N], int pos)
{
	int q = pos / 32;
	int r = pos & 31;
	for (int j = 0; j < N; j++) {
		if (q != j) {
			if (buf[j]) return false;
		} else {
			if (buf[j] != (1UL << r)) return false;
		}
	}
	return true;
}

template<int N>
static bool verifyAllZero(const uint32_t (&buf)[N])
{
	return std::count(buf, buf + N, 0) == N;
}

#endif
