#ifndef MIE_ATOI_HPP_
#define MIE_ATOI_HPP_
/**
	@file
	@brief smart atoi
	@author $Author: shigeo $
	@version $Revision: 1.5 $
	@date $Date: 2009/12/29 07:12:45 $
*/
#include <memory.h>
#include <stdexcept>
#if defined(_MSC_VER) && (_MSC_VER < 1600)
	typedef __int64 int64_t;
	typedef unsigned __int64 uint64_t;
#else
	#include <stdint.h>
#endif

namespace mie {

namespace local {

static inline void SkipZero(const char *&p, size_t& size)
{
	while (size > 0 && *p == '0') {
		p++;
		size--;
	}
}

template<typename T>
inline T Convert(bool *b, const char *p, size_t size, T overflow1, T overflow2)
{
	T x = 0;
	for (;;) {
		char c = *p;
		if (size == 0 || c == '\0') {
			*b = true;
			return x;
		}
		unsigned int y = c - '0';
		if (y <= 9) {
			if (x > overflow1 || (x == overflow1 && c >= overflow2)) {
				break;
			}
			x *= 10;
			x += y;
		} else {
			break;
		}
		p++;
		size--;
	}
	return 0;
}

template<typename T, size_t n>
T convertToInt(bool *b, const char *p, size_t size, const char (&max)[n], T min, T overflow1, char overflow2)
{
	bool isOK = false;
	T x = 0;
	bool isMinus = false;
	if (*p == '-') {
		isMinus = true;
		p++;
		size--;
	}
	if (size > 0 && *p) {
		SkipZero(p, size);
		// check minimum
		if (isMinus && size >= n - 1 && memcmp(max, p, n - 1) == 0) {
			return min;
		}
		x = Convert<T>(&isOK, p, size, overflow1, overflow2);
	}
	if (b) *b = isOK;
	if (b || isOK) return isMinus ? -x : x;
	throw std::invalid_argument("");
}

template<typename T>
T convertToUint(bool *b, const char *p, size_t size, T overflow1, char overflow2)
{
	bool isOK = false;
	T x = 0;
	if (size > 0 && *p) {
		SkipZero(p, size);
		x = Convert<T>(&isOK, p, size, overflow1, overflow2);
	}
	if (b) *b = isOK;
	if (b || isOK) return x;
	throw std::invalid_argument("");
}

} // end of local

/**
	auto detect return value class
	@note if you set bool pointer p then throw nothing and set *p = false if bad string
*/
class atoi {
	const char *p_;
	size_t size_;
	bool *b_;
	void set(bool *b, const char *p, size_t size)
	{
		b_ = b;
		p_ = p;
		size_ = size;
	}
public:
	atoi(const char *p, size_t size = -1)
	{
		set(0, p, size);
	}
	atoi(bool *b, const char *p, size_t size = -1)
	{
		set(b, p, size);
	}
	atoi(const std::string& str)
	{
		set(0, str.c_str(), str.size());
	}
	atoi(bool *b, const std::string& str)
	{
		set(b, str.c_str(), str.size());
	}

	inline operator short() const
	{
		return local::convertToInt<short>(b_, p_, size_, "32768", -32768, 3276, '8');
	}
	inline operator unsigned short() const
	{
		return local::convertToUint<unsigned short>(b_, p_, size_, 6553, '6');
	}

	inline operator int() const
	{
		return local::convertToInt<int>(b_, p_, size_, "2147483648", -2147483647 - 1, 214748364, '8');
	}

	inline operator int64_t() const
	{
		return local::convertToInt<int64_t>(b_, p_, size_, "9223372036854775808", -9223372036854775807LL -1, 922337203685477580LL, '8');
	}

	inline operator unsigned int() const
	{
		return local::convertToUint<unsigned int>(b_, p_, size_, 429496729, '6');
	}

	inline operator uint64_t() const
	{
		return local::convertToUint<uint64_t>(b_, p_, size_, 1844674407370955161ULL, '6');
	}
};

} // end of mie

#endif
