#ifndef MIE_COMMON_H_
#define MIE_COMMON_H_
/**
	@file mie_common.h
	@brief common definition
	@author $Author: shigeo $
*/

#include <stdlib.h>
#if !defined(_MSC_VER) || (_MSC_VER >= 1600)
	#include <stdint.h>
#else
	typedef unsigned __int64 uint64_t;
	typedef __int64 int64_t;
	typedef unsigned int uint32_t;
	typedef int int32_t;
	typedef unsigned short uint16_t;
	typedef signed short int16_t;
	typedef unsigned char uint8_t;
	typedef signed char int8_t;
#endif

#if defined(_MSC_VER) && !defined(MIE_SSIZE_T_IS_DEFINED)
	#define MIE_SSIZE_T_IS_DEFINED
	#ifdef __WIN64__
		typedef int64_t ssize_t;
	#else
		typedef int32_t ssize_t;
	#endif
#endif

#define MIE_LITTLE_ENDIAN
#define MIE_NUM_OF_ARRAY(x) (sizeof(x) / sizeof(*x))

namespace mie {

template<class T, size_t N>
size_t NumOfArray(const T (&)[N]) { return N; }

} // end of mie

#ifdef _MSC_VER
	#define MIE_ATOI64(x) _atoi64(x)
	#define MIE_I64 "I64"
	#ifndef MIE_ALIGN
		#define MIE_ALIGN(x) __declspec(align(x))
	#endif
#else
	#define MIE_ATOI64(x) atoll(x)
	#define MIE_I64 "ll"
	#ifndef MIE_ALIGN
		#define MIE_ALIGN(x) __attribute__((aligned(x)))
	#endif
#endif

#ifdef _MSC_VER
	#ifdef __INTEL_COMPILER // __ICL
		#define MIE_COMPILER_ICC
	#endif
	#if _MSC_VER <= 1200
		#define MIE_COMPILER_VC
		#define MIE_COMPILER_VC6
	#else
		#define MIE_COMPILER_VC
		#define MIE_COMPILER_VC7
	#endif
#endif

#ifdef _MSC_VER
	#define MIE_SNPRINTF _snprintf_s
	#define MIE_FORCE_INLINE __forceinline
	#define MIE_NOT_HERE __assume(0)
#else
	#define MIE_SNPRINTF snprintf
	#define MIE_FORCE_INLINE __attribute__((always_inline))
	#define MIE_NOT_HERE
#endif

#ifdef _MSC_VER
	#pragma warning(disable : 4514) /* remove inline */
	#pragma warning(disable : 4711) /* auto inline */
	#pragma warning(disable : 4786) /* too long identifier */
	#pragma warning(disable : 4503) /* too long name */
	#ifdef MIE_COMPILER_VC6
		#pragma warning(disable : 4127) /* constatn condition(for "if" trick) */
	#endif
	#if !defined(NOMINMAX) && defined(__cplusplus)
		#pragma message("you should define NOMINMAX for std::min, std::max")
	#endif
#endif

#ifndef MIE_DEFINED_ALLOCA
	#define MIE_DEFINED_ALLOCA
	#if defined(_MSC_VER)
		#define MIE_ALLOCA_(x) _alloca(x)
	#elif defined(__GNUC__)
		#define MIE_ALLOCA_(x) __builtin_alloca(x)
	#else
		#error "define MIE_ALLOCA_"
	#endif
	#define MIE_ALLOCA(x) (void *)(((int)MIE_ALLOCA_((x) + 16) + 15) & ~15)
	#if defined(WIN32)
		#include <malloc.h>
	#endif
#endif

#endif /* MIE_COMMON_H_ */
