#ifndef MIE_THREAD_H_
#define MIE_THREAD_H_
/**
	@file mie_thread.h
	@brief Xbh̊ȈՃNX
	@author $Author: shigeo $
	@note LinuxłłCriticalSection == mutex
*/

#include "mie_common.h"
#include <assert.h>
#ifdef _WIN32
	#include <windows.h>
	#include <process.h>
	#include <intrin.h>
	#pragma warning(disable : 4127)
#else
	#include <unistd.h>
	#include <pthread.h>
	#include <sys/time.h> // for gettimeofday
	#include <stdio.h> // for fprintf
	#include <errno.h>
#endif

namespace mie {

/**
	atomic operation
	see http://gcc.gnu.org/onlinedocs/gcc-4.4.0/gcc/Atomic-Builtins.html
*/
/**
	tmp = *p;
	*p += y;
	return tmp;
*/
static inline int AtomicAdd(int *p, int y)
{
#ifdef _WIN32
	return _InterlockedExchangeAdd((long*)p, y);
#else
	return __sync_fetch_and_add(p, y);
#endif
}

static inline unsigned int AtomicAdd(unsigned int *p, unsigned int y)
{
	return (unsigned int)AtomicAdd((int*)p, (int)y);
}

#if defined(_WIN64) || defined(__x86_64__)
static inline int64_t AtomicAdd(int64_t *p, int64_t y)
{
#ifdef _WIN32
	return _InterlockedExchangeAdd64(p, y);
#else
	return __sync_fetch_and_add(p, y);
#endif
}

static inline uint64_t AtomicAdd(uint64_t *p, uint64_t y)
{
	return (uint64_t)AtomicAdd((int64_t*)p, (int64_t)y);
}
#endif

/**
	tmp = *p;
	if (*p == oldValue) *p = newValue;
	return tmp;
*/
static inline int AtomicCompareExchange(int *p, int newValue, int oldValue)
{
#ifdef _WIN32
	return InterlockedCompareExchange((long*)p, newValue, oldValue);
#else
	return __sync_val_compare_and_swap(p, oldValue, newValue);
#endif
}

static inline int64_t AtomicCompareExchange(int64_t *p, int64_t newValue, int64_t oldValue)
{
#ifdef _WIN32
	return _InterlockedCompareExchange64(p, newValue, oldValue);
#else
	return __sync_val_compare_and_swap(p, oldValue, newValue);
#endif
}

static inline unsigned int AtomicCompareExchange(unsigned int *p, unsigned int newValue, unsigned int oldValue)
{
	return (unsigned int)AtomicCompareExchange((int*)p, (int)newValue, (int)oldValue);
}

static inline uint64_t AtomicCompareExchange(uint64_t *p, uint64_t newValue, uint64_t oldValue)
{
	return (uint64_t)AtomicCompareExchange((int64_t*)p, (int64_t)newValue, (int64_t)oldValue);
}

/**
	tmp *p;
	*p = newValue;
	return tmp;
*/
static inline int AtomicExchange(int *p, int newValue)
{
#ifdef _WIN32
	return InterlockedExchange((long*)p, newValue);
#else
	return __sync_lock_test_and_set(p, newValue);
#endif
}

static inline unsigned int AtomicExchange(unsigned int *p, unsigned int newValue)
{
	return (unsigned int)AtomicExchange((int*)p, (int)newValue);
}

static inline size_t AtomicExchangeSize_t(size_t *p, size_t newValue)
{
#ifdef _WIN32
	#ifdef _WIN64
		return (size_t)_InterlockedExchange64((__int64*)p, (__int64)newValue);
	#else
		return AtomicExchange((int*)p, (int)newValue);
	#endif
#else
	return __sync_lock_test_and_set(p, newValue);
#endif
}

/*
	ċAmutex̑
	ʏ̓IuWFNgbNXbhēxbNꍇCubN
	邪C̃tOtrueɂƁCJE^₵Ďg̓X[D
*/
const bool isRecursiveMutex = false;

#ifdef _WIN32

	#ifndef _MT
		#error "Compile under multithread mode runtime"
	#endif

	typedef HANDLE mie_thread_t;
	#define MIE_INFINITE INFINITE
	typedef unsigned __stdcall MIE_THREAD_ENTRY_TYPE(void*);

	/**
		Xbhnh𐶐
		@param hdl [out] Xbhnh
		@param entryFct [in] XbhJnGg
		@param arg [in] [Up[^
	*/
	static inline bool MIE_THREAD_Begin(mie_thread_t& hdl, MIE_THREAD_ENTRY_TYPE entryFct, void *arg, int stackSize = 0)
	{
		unsigned int threadId = 0;
		hdl = reinterpret_cast<mie_thread_t>(_beginthreadex(0, stackSize, entryFct, arg, 0, &threadId));
		return hdl != 0;
	}

	static inline bool MIE_THREAD_DETACH(mie_thread_t hdl)
	{
		int ret = CloseHandle(hdl);
		return ret != 0;
	}

	/**
		XbhI҂
		@param hdl [in] Xbhnh
		@param timeoutMsec [in] I҂(msec)
	*/
	static inline bool MIE_THREAD_Join(mie_thread_t hdl, int timeoutMsec)
	{
		bool waitRet = WaitForSingleObject(hdl, timeoutMsec) == WAIT_OBJECT_0;
		bool closeRet = CloseHandle(hdl) != 0;
		return waitRet && closeRet;
	}

	/**
		ԑ҂
		@param [in] msec ҂(msec)
	*/
	static inline void MIE_Sleep(int msec)
	{
		Sleep(msec);
	}

	typedef CRITICAL_SECTION mie_cs_t;
	static inline void MIE_CS_Init(mie_cs_t& cs) { InitializeCriticalSection(&cs); }
	static inline void MIE_CS_Lock(mie_cs_t& cs) { EnterCriticalSection(&cs); }
	static inline void MIE_CS_Unlock(mie_cs_t& cs) { LeaveCriticalSection(&cs); }
	static inline void MIE_CS_Term(mie_cs_t& cs) { DeleteCriticalSection(&cs); }

	#define mie_mutex_t HANDLE
	static inline void MIE_MUTEX_Init(mie_mutex_t& mutex)
	{
		if (isRecursiveMutex) {
			mutex = CreateMutex(NULL, FALSE, NULL);
		} else {
			mutex = CreateSemaphore(NULL /* no security */, 1 /* l */, 0x7FFFFFFF /* ől */, NULL /* OȂ */);
		}
	}
	static inline void MIE_MUTEX_Lock(mie_mutex_t& mutex) { WaitForSingleObject(mutex, INFINITE); }
	/*
		^CAEgăbNɎsfalseԂ
		@param msec [in] msec
	*/
	static inline bool MIE_MUTEX_LockTimeout(mie_mutex_t& mutex, int msec)
	{
		DWORD ret = WaitForSingleObject(mutex, msec);
		if (ret == WAIT_OBJECT_0) {
			return true;
		}
		if (ret == WAIT_TIMEOUT) {
			return false;
		}
		/* ret == WAIT_ABANDONED */
		assert(0);
		return false;
	}
	static inline void MIE_MUTEX_Unlock(mie_mutex_t& mutex)
	{
		if (isRecursiveMutex) {
			ReleaseMutex(mutex);
		} else {
			ReleaseSemaphore(mutex, 1, NULL);
		}
	}
	static inline void MIE_MUTEX_Term(mie_mutex_t& mutex) { CloseHandle(mutex); }

	class Event {
		HANDLE event_;
	public:
		Event()
		{
			event_ = ::CreateEvent(NULL, FALSE, FALSE, NULL);
		}
		~Event()
		{
			if (event_) ::CloseHandle(event_);
		}
		inline void wakeup()
		{
			if (event_) SetEvent(event_);
		}
		inline bool wait(int msec = INFINITE)
		{
			if (event_) {
				return WaitForSingleObject(event_, msec) == WAIT_OBJECT_0;
			}
			return false;
		}
	};

#elif defined(__GNUC__) || defined(__CYGWIN__)

	typedef pthread_t mie_thread_t;
	#define MIE_INFINITE 0x7fffffff
	typedef void* MIE_THREAD_ENTRY_TYPE(void*);

	/**
		Xbhnh𐶐
		@param hdl [out] Xbhnh
		@param entryFct [in] XbhJnGg
		@param arg [in] [Up[^
		@retval 0 I
	*/
	static inline bool MIE_THREAD_Begin(mie_thread_t& hdl, MIE_THREAD_ENTRY_TYPE entryFct, void *arg, int stackSize)
	{
		pthread_attr_t attr;
		if (pthread_attr_init(&attr)) return false;
		if (stackSize) if (pthread_attr_setstacksize(&attr, stackSize)) return false;
		return pthread_create(&hdl, &attr, entryFct, arg) != 0;
	}

	static inline bool MIE_THREAD_DETACH(mie_thread_t hdl)
	{
		int ret = pthread_detach(hdl);
		return ret == 0;
	}

	/**
		XbhI҂
		@param hdl [in] Xbhnh
		@param timeoutMsec [in] I҂(msec)(Linuxł͂̃p[^)
	*/
	static inline bool MIE_THREAD_Join(mie_thread_t hdl, int timeoutMsec = MIE_INFINITE)
	{
		if (timeoutMsec != MIE_INFINITE) {
			fprintf(stderr, "not support timeout\n");
		}
		return pthread_join(hdl, NULL) == 0;
	}

	/**
		ԑ҂
		@param [in] msec ҂(msec)
	*/
	static inline void MIE_Sleep(int msec)
	{
		struct timespec req;
		req.tv_sec = msec / 1000;
		req.tv_nsec = (msec % 1000) * 1000000;
		nanosleep(&req, 0);
	}

	typedef pthread_mutex_t mie_cs_t;
	static inline void MIE_CS_Init(mie_cs_t& cs) { pthread_mutex_init(&cs, NULL); }
	static inline void MIE_CS_Lock(mie_cs_t& cs) { pthread_mutex_lock(&cs); }
	static inline void MIE_CS_Unlock(mie_cs_t& cs) { pthread_mutex_unlock(&cs); }
	static inline void MIE_CS_Term(mie_cs_t& cs) { pthread_mutex_destroy(&cs); }

	typedef pthread_mutex_t mie_mutex_t;
	static inline void MIE_MUTEX_Init(mie_mutex_t& mutex)
	{
		if (isRecursiveMutex) {
			pthread_mutexattr_t mutexattr;
			pthread_mutexattr_init(&mutexattr);
			/* pthreadW */
#ifndef PTHREAD_MUTEX_RECURSIVE_NP
			pthread_mutexattr_settype(&mutexattr, PTHREAD_MUTEX_RECURSIVE);
#else
			pthread_mutexattr_settype(&mutexattr, PTHREAD_MUTEX_RECURSIVE_NP);
#endif
			pthread_mutex_init(&mutex, &mutexattr);
		} else {
			pthread_mutex_init(&mutex, NULL);
		}
	}
	static inline void MIE_MUTEX_Lock(mie_mutex_t& mutex) { pthread_mutex_lock(&mutex); }
	static inline bool MIE_MUTEX_LockTimeout(mie_mutex_t& mutex, int msec)
	{
		timespec absTime;
		clock_gettime(CLOCK_REALTIME, &absTime);
		absTime.tv_sec += msec / 1000;
		absTime.tv_nsec += msec % 1000;
		bool ret = pthread_mutex_timedlock(&mutex, &absTime) == 0;
//		fprintf(stderr, "QQQ err=%d, %s\n", errno, strerror_r(errno, 0, 0));
		return ret;
	}
	static inline void MIE_MUTEX_Unlock(mie_mutex_t& mutex) { pthread_mutex_unlock(&mutex); }
	static inline void MIE_MUTEX_Term(mie_mutex_t& mutex) { pthread_mutex_destroy(&mutex); }

	typedef pthread_cond_t mie_cond_t;
	static inline void MIE_COND_Init(mie_cond_t *cond) { pthread_cond_init(cond, 0); }
	static inline void MIE_COND_Notify(mie_cond_t *cond) { pthread_cond_signal(cond); }
	static inline bool MIE_COND_Wait(mie_cond_t *cond, mie_mutex_t *mutex) { return pthread_cond_wait(cond, mutex) == 0; }
	static inline bool MIE_COND_TimedWait(mie_cond_t *cond, mie_mutex_t *mutex, int waitMsec) {
		const int oneSec = 1000 * 1000 * 1000;
		struct timespec ts;
		struct timeval now;
		gettimeofday(&now, 0);
		ts.tv_sec = now.tv_sec;
		ts.tv_nsec = (now.tv_usec + waitMsec * 1000) * 1000;
		if (ts.tv_nsec >= oneSec) {
			ts.tv_sec += ts.tv_nsec / oneSec;
			ts.tv_nsec %= oneSec;
		}
		int ret;
		do {
			ret = pthread_cond_timedwait(cond, mutex, &ts);
		} while (ret == EINTR);
		if (ret == 0) return true;
		if (ret == ETIMEDOUT) return false;
		assert(0);
		return false;
	}
	static inline void MIE_COND_Term(mie_cond_t *cond) { pthread_cond_destroy(cond); }

#else
	#error "not supported"
#endif

/**
	XbhNX
*/
template<class Derived>
class ThreadBase {
private:
    ThreadBase(const ThreadBase&);
    void operator=(const ThreadBase&);
protected:
	mie_thread_t threadHdl_;

	/*
		Xbh[vC^tF[X
	*/
	static void threadLoopIF(void *arg)
	{
		Derived* main = reinterpret_cast<Derived*>(arg);
		main->threadEntry();
		// end of thread
//		main->detachThread();
	}
public:
	/**
		RXgN^
	*/
	ThreadBase()
		: threadHdl_(0)
	{
	}
	~ThreadBase()
	{
		joinThread();
	}

	/**
		XbhJn
	*/
	bool beginThread(int stackSize = 0)
	{
		assert(threadHdl_ == 0);
		return MIE_THREAD_Begin(threadHdl_, (MIE_THREAD_ENTRY_TYPE*)threadLoopIF, &static_cast<Derived&>(*this), stackSize);
	}

#if 0
	/**
		Xbh~j
	*/
	bool termThread(int waitMsec = -1)
	{
		if (threadHdl_) {
			static_cast<Derived&>(*this).stopThread();
			if (!joinThread(waitMsec)) {
				return false;
			}
			MIE_THREAD_DETACH(&threadHdl_);
		}
		return true;
	}
#endif
	/**
		detach
	*/
	bool detachThread()
	{
		mie_thread_t hdl = (mie_thread_t)AtomicExchangeSize_t((size_t*)&threadHdl_, 0);
		if (hdl) {
			return MIE_THREAD_DETACH(hdl);
		}
		return false;
	}

	/**
		Xbh̏I҂
		@param waitMsec [in] ^CAEg(ftHg)
	*/
	bool joinThread(int waitMsec = MIE_INFINITE)
	{
		mie_thread_t hdl = (mie_thread_t)AtomicExchangeSize_t((size_t*)&threadHdl_, 0);
		if (hdl) {
			if (MIE_THREAD_Join(hdl, waitMsec)) {
				return false;
			}
		}
		return true;
	}
};

namespace Thread {
/**
	Xbh̗D揇
*/
enum Priority {
	HIGHEST, /* łD揇 */
	HIGH, /* D揇 */
	NORMAL, /* ʏ */
	LOW /* Ⴂ */
};

/**
	Xbh̗D揇ʂύX
	@param priority [in] D揇
*/
static inline bool SetPriorityOfThread(Thread::Priority priority)
{
#ifdef WIN32
	int priorityValue;
	HANDLE hdl = GetCurrentThread();
	switch (priority) {
	case HIGHEST:
		priorityValue = THREAD_PRIORITY_HIGHEST;
	case HIGH:
		priorityValue = THREAD_PRIORITY_ABOVE_NORMAL;
		break;
	case NORMAL:
		priorityValue = THREAD_PRIORITY_NORMAL;
		break;
	case LOW:
		priorityValue = THREAD_PRIORITY_BELOW_NORMAL;
		break;
	default:
		return false;
	}
	::SetThreadPriority(hdl, priorityValue);
#else
	fprintf(stderr, "not support setThreadPriority(%d)\n", priority);
#endif
	return true;
}
/**
	vZX̗D揇ʂύX
*/
static inline bool SetPriorityOfProcess(Thread::Priority priority)
{
#ifdef WIN32
	DWORD priorityClass;
	HANDLE hdl = GetCurrentProcess();
	switch (priority) {
	case HIGHEST:
		priorityClass = HIGH_PRIORITY_CLASS;
	case HIGH:
		priorityClass = ABOVE_NORMAL_PRIORITY_CLASS;
		break;
	case NORMAL:
		priorityClass = NORMAL_PRIORITY_CLASS;
		break;
	case LOW:
		priorityClass = BELOW_NORMAL_PRIORITY_CLASS;
		break;
	default:
		return false;
	}
	SetPriorityClass(hdl, priorityClass);
#else
	fprintf(stderr, "not support setProcessPriority(%d)\n", priority);
#endif
	return true;
}

/**
	ݎ擾
	@return oߎ(msec)
	@note ݂̂Ӗ(I[o[t[邩Ȃ)
*/
static inline unsigned int GetCurrTimeMsec()
{
#ifdef _MSC_VER
	return GetTickCount();
#else
	struct timeval t;
	gettimeofday(&t, 0);
	/* 1000000000sec = 31 years */
	return (t.tv_sec - 1000000000) * 1000 + (t.tv_usec / 1000);
#endif
}


} // end of mie::Thread

/**
	yrNX
*/
class CriticalSection {
public:
	CriticalSection()
	{
		MIE_CS_Init(cs_);
	}
	~CriticalSection()
	{
		MIE_CS_Term(cs_);
	}
	inline void lock()
	{
		MIE_CS_Lock(cs_);
	}
	inline void unlock()
	{
		MIE_CS_Unlock(cs_);
	}
	inline mie_cs_t *get()
	{
		return &cs_;
	}
private:
	CriticalSection(const CriticalSection&); /* Rs[Ȃ */
	CriticalSection& operator=(const CriticalSection&);
	mie_cs_t cs_;
};

/**
	mutex(^CAEg)
*/
class Mutex {
public:
	Mutex()
	{
		MIE_MUTEX_Init(mutex_);
	}
	~Mutex()
	{
		MIE_MUTEX_Term(mutex_);
	}
	inline void lock()
	{
		MIE_MUTEX_Lock(mutex_);
	}
	inline bool lockTimeout(int msec)
	{
		return MIE_MUTEX_LockTimeout(mutex_, msec);
	}
	inline void unlock()
	{
		MIE_MUTEX_Unlock(mutex_);
	}
	inline mie_mutex_t *get()
	{
		return &mutex_;
	}
private:
	Mutex(const Mutex&); /* Rs[Ȃ */
	Mutex& operator=(const Mutex&);
	mie_mutex_t mutex_;
};

/**
	bN
*/
template<class T>
class AutoLockT {
public:
	AutoLockT(T &cs)
		: cs_(cs)
	{
		cs_.lock();
	}
	~AutoLockT()
	{
		cs_.unlock();
	}
private:
	T& cs_;
	AutoLockT& operator=(const AutoLockT&);
};

#ifdef __GNUC__
class CondVar {
public:
	CondVar()
	{
		MIE_COND_Init(&cond_);
	}
	~CondVar()
	{
		MIE_COND_Term(&cond_);
	}
	inline void notify()
	{
		MIE_COND_Notify(&cond_);
	}
	inline bool wait(mie::Mutex& mutex)
	{
		return MIE_COND_Wait(&cond_, mutex.get());
	}
	inline bool waitTimeout(mie::Mutex& mutex, int waitMsec)
	{
		return MIE_COND_TimedWait(&cond_, mutex.get(), waitMsec);
	}
private:
	CondVar(const CondVar&); /* don't copy */
	CondVar& operator=(const CondVar&);
	mie_cond_t cond_;
};

class Event {
	mie::Mutex mutex_;
	Event(const Event&);
	void operator=(const Event&);
public:
	Event()
	{
		mutex_.lock();
	}
	~Event()
	{
		mutex_.unlock();
	}
	inline void wakeup()
	{
		mutex_.unlock();
	}
	inline bool wait(int msec = 3600 * 1000)
	{
#if 0
		mutex_.lock();
		return true;
#else
		if (mutex_.lockTimeout(msec)) {
			return true;
		}
		return false;
#endif
	}
};
#endif

typedef AutoLockT<CriticalSection> AutoLock; /*!< auto lock critical section */

/**
	bN(^CAEg)
*/
class AutoLockTimeout {
public:
	AutoLockTimeout(Mutex& mutex, int msec = MIE_INFINITE)
		: mutex_(mutex)
	{
		if (msec == MIE_INFINITE) {
			isLocked_ = true;
			mutex_.lock();
		} else {
			isLocked_ = mutex_.lockTimeout(msec);
		}
	}
	~AutoLockTimeout()
	{
		if (isLocked_) mutex_.unlock();
	}
	bool isLocked() const { return isLocked_; }
private:
	Mutex& mutex_;
	bool isLocked_;
	AutoLockTimeout& operator=(const AutoLockTimeout&);
};

} // end of namespace

#endif /* MIE_THREAD_H_ */

